# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Iterator
from unittest import mock

import grpc
import pytest
from buildgrid_metering.client.exceptions import MeteringServiceHTTPError
from flaky import flaky
from google.protobuf.any_pb2 import Any as ProtoAny

from buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2_grpc import add_PushServicer_to_server
from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    Action,
    Digest,
    ExecuteOperationMetadata,
    ExecuteResponse,
)
from buildgrid._protos.build.buildbox.execution_stats_pb2 import ExecutionStatistics
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.actioncache.caches.lru_cache import LruActionCache
from buildgrid.server.cas.storage import lru_memory_cache
from buildgrid.server.cas.storage.storage_abc import create_write_session
from buildgrid.server.client.asset import AssetClient
from buildgrid.server.context import instance_context
from buildgrid.server.controller import ExecutionController
from buildgrid.server.enums import BotStatus, LeaseState, OperationStage
from buildgrid.server.exceptions import InvalidArgumentError, NotFoundError, ResourceExhaustedError, StorageFullError
from buildgrid.server.scheduler import PropertySet, Scheduler
from buildgrid.server.scheduler.properties import hash_from_dict
from buildgrid.server.sql.models import (
    ClientIdentityEntry,
    JobEntry,
    OperationEntry,
    digest_to_string,
    string_to_digest,
)
from buildgrid.server.sql.provider import SqlProvider
from buildgrid.server.utils.digests import create_digest
from tests.utils.fixtures import mock_logstream_client
from tests.utils.server import MockAssetPushServicer
from tests.utils.utils import find_free_port

server = mock.create_autospec(grpc.Server)

command = remote_execution_pb2.Command()
command_digest = create_digest(command.SerializeToString())

action = remote_execution_pb2.Action(command_digest=command_digest, do_not_cache=False)
action_digest = create_digest(action.SerializeToString())

uncacheable_action = remote_execution_pb2.Action(command_digest=command_digest, do_not_cache=True)
uncacheable_action_digest = create_digest(uncacheable_action.SerializeToString())


@pytest.fixture
def asset_service_remote() -> str:
    return f"localhost:{find_free_port()}"


@pytest.fixture
def push_service(asset_service_remote: str) -> Iterator[MockAssetPushServicer]:
    server = grpc.server(thread_pool=ThreadPoolExecutor())
    push = MockAssetPushServicer()
    add_PushServicer_to_server(push, server)
    server.add_insecure_port(asset_service_remote)
    server.start()

    try:
        yield push
    finally:
        server.stop(grace=None)


@pytest.fixture
def asset_service_channel(asset_service_remote: str) -> Iterator[grpc.Channel]:
    with grpc.insecure_channel(asset_service_remote) as channel:
        yield channel


@pytest.fixture
def asset_client(push_service, asset_service_channel: str) -> Iterator[AssetClient]:
    yield AssetClient(asset_service_channel, instance_name="asset")


PARAMS = ["action-cache", "no-action-cache"]


@pytest.fixture(params=PARAMS)
def controller(request, asset_client: AssetClient, common_props: PropertySet):
    use_cache = request.param
    storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
    db = tempfile.NamedTemporaryFile().name
    sql_provider = SqlProvider(connection_string=f"sqlite:///{db}", automigrate=True)

    with create_write_session(command_digest) as write_session:
        write_session.write(command.SerializeToString())
        storage.commit_write(command_digest, write_session)

    with create_write_session(action_digest) as write_session:
        write_session.write(action.SerializeToString())
        storage.commit_write(action_digest, write_session)

    if use_cache == "action-cache":
        cache = LruActionCache(storage, 50)
        scheduler = Scheduler(
            sql_provider,
            storage,
            poll_interval=0.01,
            action_cache=cache,
            property_set=common_props,
            asset_client=asset_client,
            queued_action_retention_hours=12,
            completed_action_retention_hours=1,
            action_result_retention_hours=3,
        )
    else:
        scheduler = Scheduler(
            sql_provider,
            storage,
            poll_interval=0.01,
            property_set=common_props,
            asset_client=asset_client,
            queued_action_retention_hours=12,
            completed_action_retention_hours=1,
            action_result_retention_hours=3,
        )

    with scheduler, instance_context("sql"):
        yield ExecutionController(scheduler)


PARAMS_MAX_EXECUTION = [None, 0.1, 2]


@pytest.fixture(params=PARAMS_MAX_EXECUTION)
def controller_max_execution_timeout(request, common_props):
    max_execution_timeout = request.param
    storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
    db = tempfile.NamedTemporaryFile().name
    sql_provider = SqlProvider(connection_string=f"sqlite:///{db}", automigrate=True)

    with create_write_session(command_digest) as write_session:
        write_session.write(command.SerializeToString())
        storage.commit_write(command_digest, write_session)

    with create_write_session(action_digest) as write_session:
        write_session.write(action.SerializeToString())
        storage.commit_write(action_digest, write_session)

    scheduler = Scheduler(
        sql_provider,
        storage,
        poll_interval=0.1,
        property_set=common_props,
        max_execution_timeout=max_execution_timeout,
    )
    controller = ExecutionController(scheduler)
    with controller.execution_instance, instance_context("sql"):
        yield controller


PARAMS_MAX_QUEUE_SIZE = [None, 1, 5]


@pytest.fixture(params=PARAMS_MAX_QUEUE_SIZE)
def controller_max_queue_size(request, common_props):
    max_queue_size = request.param
    storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
    db = tempfile.NamedTemporaryFile().name
    sql_provider = SqlProvider(connection_string=f"sqlite:///{db}", automigrate=True)

    with create_write_session(command_digest) as write_session:
        write_session.write(command.SerializeToString())
        storage.commit_write(command_digest, write_session)

    with create_write_session(action_digest) as write_session:
        write_session.write(action.SerializeToString())
        storage.commit_write(action_digest, write_session)

    scheduler = Scheduler(
        sql_provider,
        storage,
        poll_interval=0.1,
        property_set=common_props,
        max_queue_size=max_queue_size,
    )
    controller = ExecutionController(scheduler)
    with controller.execution_instance, instance_context("sql"):
        yield controller


@pytest.fixture
def isolated_controllers(request, common_props):
    storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
    db = tempfile.NamedTemporaryFile().name
    sql_provider = SqlProvider(connection_string=f"sqlite:///{db}", automigrate=True)

    with create_write_session(command_digest) as write_session:
        write_session.write(command.SerializeToString())
        storage.commit_write(command_digest, write_session)

    with create_write_session(action_digest) as write_session:
        write_session.write(action.SerializeToString())
        storage.commit_write(action_digest, write_session)

    cache = LruActionCache(storage, 50)
    scheduler_a = Scheduler(sql_provider, storage, action_cache=cache, property_set=common_props)
    scheduler_b = Scheduler(sql_provider, storage, action_cache=cache, property_set=common_props)
    with scheduler_b, scheduler_a:
        yield ExecutionController(scheduler_a), ExecutionController(scheduler_b)


def mock_queue_job_action(scheduler, skip_cache_lookup=True, request_metadata=None, do_not_cache=False, assign=True):
    bot_id = "test-worker"
    bot_name = scheduler.add_bot_entry(bot_session_id=bot_id, bot_session_status=BotStatus.OK.value)
    operation_name = scheduler.queue_job_action(
        action=uncacheable_action if do_not_cache else action,
        action_digest=uncacheable_action_digest if do_not_cache else action_digest,
        command=command,
        platform_requirements={},
        property_label="unknown",
        priority=0,
        skip_cache_lookup=skip_cache_lookup,
        request_metadata=request_metadata,
    )
    if assign:
        scheduler.assign_n_leases_by_priority(capability_hash=hash_from_dict({}), bot_names=[bot_name])
    job_name = scheduler.get_operation_job_name(operation_name)
    return bot_name, bot_id, operation_name, job_name


@pytest.mark.parametrize("monitoring", [True, False])
def test_update_lease_state(controller, monitoring, push_service: MockAssetPushServicer):
    scheduler = controller.execution_instance.scheduler

    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)
    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        # check retention for queued action
        assert job.action_digest in push_service.blobs
        asset = push_service.blobs[job.action_digest]
        assert asset.instance_name == "asset"
        assert asset.uris == [f"nih:sha-256;{string_to_digest(job.action_digest).hash}"]
        action = Action.FromString(job.action)
        assert action.command_digest in asset.references_blobs
        assert action.input_root_digest in asset.references_directories
        assert asset.expire_at.ToDatetime() > datetime.datetime.now() + datetime.timedelta(hours=10)

    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
    assert lease.state == LeaseState.PENDING.value

    lease.state = LeaseState.ACTIVE.value
    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
    assert lease.state == LeaseState.ACTIVE.value

    mock_stderr_digest = Digest(hash="stderr", size_bytes=42)
    test_action_result = remote_execution_pb2.ActionResult(
        exit_code=0, stdout_raw=b"test", stderr_digest=mock_stderr_digest
    )
    lease.result.Pack(test_action_result)
    lease.state = LeaseState.COMPLETED.value
    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
    assert lease is None
    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        # check retention for completed action and result
        asset = push_service.blobs[job.action_digest]
        assert asset.expire_at.ToDatetime() > datetime.datetime.now() + datetime.timedelta(hours=0.5)
        action_result = scheduler.storage.get_message(string_to_digest(job.result), ExecuteResponse).result
        result_digest = create_digest(action_result.SerializeToString())
        result_asset = push_service.blobs[digest_to_string(result_digest)]
        assert result_asset.instance_name == "asset"
        assert result_asset.uris == [f"nih:sha-256;{result_digest.hash}"]
        assert mock_stderr_digest in result_asset.references_blobs
        assert result_asset.expire_at.ToDatetime() > datetime.datetime.now() + datetime.timedelta(hours=2)


def test_retry_job_lease(controller):
    scheduler = controller.execution_instance.scheduler
    scheduler.max_job_attempts = 2

    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)

    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        assert job.n_tries == 1
        assert job.stage == OperationStage.QUEUED.value

    scheduler.close_bot_sessions(bot_name)
    bot_name = scheduler.add_bot_entry(bot_session_id=bot_id, bot_session_status=BotStatus.OK.value)
    scheduler.assign_n_leases_by_priority(capability_hash=hash_from_dict({}), bot_names=[bot_name])

    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        assert job.n_tries == 2
        assert job.stage == OperationStage.QUEUED.value

    scheduler.close_bot_sessions(bot_name)

    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        assert job.n_tries == 2
        assert job.stage == OperationStage.COMPLETED.value
        assert job.status_code == code_pb2.ABORTED


def test_requeue_queued_job(controller):
    scheduler = controller.execution_instance.scheduler
    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)

    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        assert job.leases[0].worker_name is not None
        assert job.stage == OperationStage.QUEUED.value

    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
    lease.state = LeaseState.ACTIVE.value
    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
    assert lease.state == LeaseState.ACTIVE.value

    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        assert job.leases[0].worker_name is not None
        assert job.stage == OperationStage.EXECUTING.value

    # Make sure that retrying a job that was assigned but
    # not marked as in progress properly re-queues
    scheduler.close_bot_sessions(bot_name)
    bot_name = scheduler.add_bot_entry(bot_session_id=bot_id, bot_session_status=BotStatus.OK.value)

    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        assert job.leases[0].worker_name is None
        assert job.stage == OperationStage.QUEUED.value

    scheduler.assign_n_leases_by_priority(capability_hash=hash_from_dict({}), bot_names=[bot_name])

    with scheduler._sql.session() as session:
        job = scheduler._get_job(job_name, session)
        assert job.leases[0].worker_name is not None
        assert job.stage == OperationStage.QUEUED.value


# Test that jobs can be created/completed with no action-cache, a working action cache,
# and an action-cache that throws exceptions when used
@pytest.mark.parametrize("cache_errors", [True, False])
def test_complete_lease_action_cache_error(controller, cache_errors):
    scheduler = controller.execution_instance.scheduler
    if scheduler.action_cache and cache_errors:
        with mock.patch.object(scheduler.action_cache, "get_action_result", autospec=True) as ac_mock:
            ac_mock.side_effect = ConnectionError("Fake Connection Error")
            bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler, skip_cache_lookup=False)
            assert ac_mock.call_count == 1
    else:
        bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)

    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
    assert lease.state == LeaseState.PENDING.value

    lease.state = LeaseState.ACTIVE.value
    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
    assert lease.state == LeaseState.ACTIVE.value

    lease.state = LeaseState.COMPLETED.value
    if scheduler.action_cache and cache_errors:
        with mock.patch.object(scheduler.action_cache, "update_action_result", autospec=True) as ac_mock:
            ac_mock.side_effect = StorageFullError("TestActionCache is full")
            lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
            assert ac_mock.call_count == 1
    else:
        lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)

    assert lease is None


def test_uncacheable_action(controller):
    scheduler = controller.execution_instance.scheduler
    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler, do_not_cache=True)

    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
    assert lease.state == LeaseState.PENDING.value

    lease.state = LeaseState.ACTIVE.value
    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
    assert lease.state == LeaseState.ACTIVE.value

    lease.state = LeaseState.COMPLETED.value
    if scheduler.action_cache:
        # update_action_result should not be called for do_not_cache actions
        with mock.patch.object(scheduler.action_cache, "update_action_result", autospec=True) as ac_mock:
            lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
            assert ac_mock.call_count == 0
    else:
        lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)

    assert lease is None


def test_action_deduplication(controller):
    # Verify two identical actions are deduplicated
    scheduler = controller.execution_instance.scheduler
    _, _, operation_name, job_name = mock_queue_job_action(scheduler)
    _, _, operation_name2, job_name2 = mock_queue_job_action(scheduler)
    # Operation names should be different, but job names identical
    assert operation_name != operation_name2
    assert job_name == job_name2

    # Actions with do_not_cache=True must not be deduplicated
    _, _, uncacheable_operation_name, uncacheable_job_name = mock_queue_job_action(scheduler, do_not_cache=True)
    _, _, uncacheable_operation_name2, uncacheable_job_name2 = mock_queue_job_action(scheduler, do_not_cache=True)
    assert uncacheable_operation_name != uncacheable_operation_name2
    assert uncacheable_job_name != uncacheable_job_name2


def test_get_metadata_for_leases(controller):
    scheduler = controller.execution_instance.scheduler

    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)

    scheduler.logstream_channel = mock.MagicMock()
    scheduler.logstream_instance = "testing"

    with mock.patch("buildgrid.server.scheduler.impl.logstream_client", new=mock_logstream_client()):
        lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
        lease.state = LeaseState.ACTIVE.value
        lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)

    lease_metadata = scheduler.get_metadata_for_leases([lease])
    op = scheduler.load_operation(operation_name)
    op_metadata = ExecuteOperationMetadata()
    op.metadata.Unpack(op_metadata)

    assert op_metadata.partial_execution_metadata.worker == bot_id
    assert op_metadata.partial_execution_metadata.queued_timestamp.ToDatetime() != datetime.datetime(1970, 1, 1, 0, 0)
    assert op_metadata.partial_execution_metadata.worker_start_timestamp.ToDatetime() != datetime.datetime(
        1970, 1, 1, 0, 0
    )
    assert op_metadata.partial_execution_metadata.worker_completed_timestamp.ToDatetime() == datetime.datetime(
        1970, 1, 1, 0, 0
    )

    assert op_metadata.stdout_stream_name
    assert op_metadata.stdout_stream_name.endswith("stdout/mock-logstream")
    assert op_metadata.stderr_stream_name
    assert op_metadata.stderr_stream_name.endswith("stderr/mock-logstream")

    # Lease metadata should be the same, just with the write streams instead
    op_metadata.stdout_stream_name = op_metadata.stdout_stream_name + "/write"
    op_metadata.stderr_stream_name = op_metadata.stderr_stream_name + "/write"
    assert lease_metadata[0][0] == "executeoperationmetadata-bin"
    assert lease_metadata[0][1] == op_metadata.SerializeToString()


def test_list_operations(controller):
    """Test that the scheduler reports the correct
    number of operations when calling list_operations()"""
    scheduler = controller.execution_instance.scheduler
    operations, _ = scheduler.list_operations()
    assert len(operations) == 0

    mock_queue_job_action(scheduler)

    operations, _ = scheduler.list_operations()
    assert len(operations) == 1

    mock_queue_job_action(scheduler)
    mock_queue_job_action(scheduler)

    operations, _ = scheduler.list_operations()
    assert len(operations) == 3


def test_query_operation_metadata(controller):
    """Test that the scheduler returns the expected `RequestMetadata`
    information for an operation.
    """
    scheduler = controller.execution_instance.scheduler

    request_metadata = remote_execution_pb2.RequestMetadata()
    request_metadata.tool_details.tool_name = "my-tool"
    request_metadata.tool_details.tool_version = "1.0"
    request_metadata.tool_invocation_id = "invId123"
    request_metadata.correlated_invocations_id = "corId456"

    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler, request_metadata=request_metadata)
    metadata = scheduler.get_operation_request_metadata_by_name(operation_name)

    assert metadata is not None
    assert metadata.tool_details.tool_name == request_metadata.tool_details.tool_name
    assert metadata.tool_details.tool_version == request_metadata.tool_details.tool_version
    assert metadata.tool_invocation_id == request_metadata.tool_invocation_id
    assert metadata.correlated_invocations_id == request_metadata.correlated_invocations_id


# Validate that an ongoing job has an operation which is not marked done
def test_get_job_operation_ongoing(controller):
    scheduler = controller.execution_instance.scheduler
    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)

    operation = scheduler.load_operation(operation_name)
    assert operation.name == operation_name
    assert not operation.done


# Validate that a finished job has an operation with an ExecuteResponse
# containing an OK status code
def test_get_job_operation_finished(controller):
    scheduler = controller.execution_instance.scheduler
    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)

    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
    lease.state = LeaseState.ACTIVE.value
    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)

    lease.state = LeaseState.COMPLETED.value
    test_action_result = remote_execution_pb2.ActionResult(exit_code=1, stdout_raw=b"test")
    lease.result.Pack(test_action_result)
    scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)

    operation = scheduler.load_operation(operation_name)
    assert operation.name == operation_name
    assert operation.done
    assert operation.WhichOneof("result") == "response"

    execute_response = remote_execution_pb2.ExecuteResponse()
    assert operation.response.Unpack(execute_response)

    assert execute_response.status.code == code_pb2.OK
    assert execute_response.result.stdout_raw == b"test"
    assert execute_response.result.exit_code == 1


# Validate that a failed job has an operation with an ExecuteResponse
# containing a non-OK status code
def test_get_job_operation_failed(controller):
    scheduler = controller.execution_instance.scheduler
    scheduler.max_job_attempts = 0
    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)

    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
    lease.state = LeaseState.ACTIVE.value
    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)

    lease.state = LeaseState.COMPLETED.value
    failed_status = Status(code=code_pb2.ABORTED, message="Failed Status")
    lease.status.CopyFrom(failed_status)
    scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)

    operation = scheduler.load_operation(operation_name)
    assert operation.name == operation_name
    assert operation.done
    assert operation.WhichOneof("result") == "response"

    execute_response = remote_execution_pb2.ExecuteResponse()
    assert operation.response.Unpack(execute_response)
    assert execute_response.status == failed_status


@flaky(max_runs=3)
@pytest.mark.parametrize("n_operations", [1, 2])
@pytest.mark.parametrize("sleep_time", [0.2])
def test_max_execution_timeout(controller_max_execution_timeout, sleep_time, n_operations):
    scheduler = controller_max_execution_timeout.execution_instance.scheduler

    max_execution_timeout = scheduler.max_execution_timeout

    # Queue Job
    bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)

    # Create n operations
    operation_names = [operation_name]
    for i in range(n_operations - 1):
        operation_names.append(scheduler.create_operation(job_name=job_name))

    for i in range(n_operations):
        operation_initially = scheduler.load_operation(operation_names[i])
        assert operation_initially.done is False

    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
    lease.state = LeaseState.ACTIVE.value
    lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)

    # Make sure it was marked as "Executing"
    op_after_assignment = scheduler.load_operation(operation_name)
    assert not op_after_assignment.done

    # Wait and see it marked as cancelled when exceeding execution timeout...
    time.sleep(sleep_time)

    if max_execution_timeout and sleep_time >= max_execution_timeout:
        for operation_name in operation_names:
            operation = scheduler.load_operation(operation_name)
            assert operation.done is True
            # Verify the response in the operation has the correct
            # status code
            assert operation.WhichOneof("result") == "response"
            execute_response = remote_execution_pb2.ExecuteResponse()
            assert operation.response.Unpack(execute_response)
            assert execute_response.status.code == code_pb2.DEADLINE_EXCEEDED

        lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
        assert lease.state == LeaseState.CANCELLED.value
        lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
        assert lease is None
    else:
        for operation_name in operation_names:
            operation = scheduler.load_operation(operation_name)
            assert operation.done is False
        lease = scheduler.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, lease)
        assert lease.state == LeaseState.ACTIVE.value


@pytest.fixture
def mock_metering_client():
    usage_sink = []

    def mock_put_usage(identity, operation_name, usage):
        usage_sink.append((identity, operation_name, usage))

    mocked = mock.Mock()
    mocked.put_usage.side_effect = mock_put_usage
    mocked.sink = usage_sink
    yield mocked


def store_message(scheduler, message):
    binary = message.SerializeToString()
    digest = create_digest(binary)
    scheduler.storage.bulk_update_blobs([(digest, binary)])
    return digest


@pytest.fixture
def mock_execution_metadata(controller):
    scheduler = controller.execution_instance.scheduler

    action_digest = store_message(scheduler, Action())

    execution_stats = ExecutionStatistics()
    execution_stats.command_rusage.maxrss = 100
    execution_stats_any = ProtoAny()
    execution_stats_any.Pack(execution_stats)
    execution_stats_digest = store_message(scheduler, execution_stats_any)

    digest_any = ProtoAny()
    digest_any.Pack(execution_stats_digest)
    execution_metadata = remote_execution_pb2.ExecutedActionMetadata()
    execution_metadata.auxiliary_metadata.append(digest_any)

    now = datetime.datetime.now()
    with scheduler._sql.session(exceptions_to_not_raise_on=[Exception]) as session:
        session.add(
            JobEntry(
                name="job",
                instance_name="sql",
                action_digest=digest_to_string(action_digest),
                action=b"",
                create_timestamp=now,
                queued_timestamp=now,
                assigned=False,
                n_tries=1,
                platform_requirements="",
                property_label="unknown",
                command="foo",
            )
        ),
        session.add(ClientIdentityEntry(id=1, instance="sql", workflow="build", actor="tool", subject="user1"))
        session.add(ClientIdentityEntry(id=2, instance="sql", workflow="build", actor="tool", subject="user2"))
        session.add(OperationEntry(name="op1", job_name="job", client_identity_id=1))
        session.add(OperationEntry(name="op2", job_name="job", client_identity_id=2))

    return execution_metadata


def test_publish_execution_stats(controller, mock_metering_client, mock_execution_metadata):
    scheduler = controller.execution_instance.scheduler
    scheduler.metering_client = mock_metering_client

    with instance_context("sql"):
        scheduler.publish_execution_stats("job", mock_execution_metadata)
    assert len(mock_metering_client.sink) == 2


def test_publish_execution_stats_fetch_fails_no_throw(controller, mock_metering_client, mock_execution_metadata):
    scheduler = controller.execution_instance.scheduler
    scheduler.metering_client = mock_metering_client
    mock_metering_client.put_usage.side_effect = MeteringServiceHTTPError(503, "big bang")

    with instance_context("sql"):
        scheduler.publish_execution_stats("job", mock_execution_metadata)
    assert len(mock_metering_client.sink) == 0


def test_publish_execution_stats_invalid_execution_stats(controller, mock_metering_client, mock_execution_metadata):
    scheduler = controller.execution_instance.scheduler
    scheduler.metering_client = mock_metering_client

    some_digest = Digest()
    some_digest_any = ProtoAny()
    some_digest_any.Pack(some_digest)
    mock_execution_metadata.auxiliary_metadata.pop()
    mock_execution_metadata.auxiliary_metadata.append(some_digest_any)

    with instance_context("sql"):
        scheduler.publish_execution_stats("job", mock_execution_metadata)
    assert len(mock_metering_client.sink) == 0


def test_publish_execution_stats_no_client(controller, mock_metering_client, mock_execution_metadata):
    scheduler = controller.execution_instance.scheduler
    # Just dont assign a client?? Weird test

    with instance_context("sql"):
        scheduler.publish_execution_stats("job", mock_execution_metadata)
    assert len(mock_metering_client.sink) == 0


def test_instance_isolation_create_operation(isolated_controllers):
    controller_a, controller_b = isolated_controllers
    scheduler_a = controller_a.execution_instance.scheduler
    scheduler_b = controller_b.execution_instance.scheduler

    with instance_context("instance-a"):
        bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler_a)
        # Check that we can only create an Operation for the job on the same instance
        scheduler_a.create_operation(job_name=job_name)

    with instance_context("instance-b"):
        with pytest.raises(NotFoundError):
            scheduler_b.create_operation(job_name=job_name)


def test_instance_isolation_load_operation(isolated_controllers):
    controller_a, controller_b = isolated_controllers
    scheduler_a = controller_a.execution_instance.scheduler
    scheduler_b = controller_b.execution_instance.scheduler

    with instance_context("instance-a"):
        bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler_a)
        # Check that fetching the operation respects instances
        scheduler_a.load_operation(operation_name)

    with instance_context("instance-b"):
        with pytest.raises(NotFoundError):
            scheduler_b.load_operation(operation_name)


def test_instance_isolation_lease_assignment(isolated_controllers):
    controller_a, controller_b = isolated_controllers
    scheduler_a = controller_a.execution_instance.scheduler
    scheduler_b = controller_b.execution_instance.scheduler

    with instance_context("instance-a"):
        bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler_a)

        lease = scheduler_a.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
        assert lease.state == LeaseState.PENDING.value

    with instance_context("instance-b"):
        with pytest.raises(InvalidArgumentError):
            scheduler_b.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)

    with instance_context("instance-a"):
        scheduler_a.close_bot_sessions(bot_name)
        bot_name = scheduler_a.add_bot_entry(bot_session_id=bot_id, bot_session_status=BotStatus.OK.value)

    # Attempt to assign leases from the instance which didn't get the job
    with instance_context("instance-b"):
        scheduler_b.assign_n_leases_by_priority(capability_hash=hash_from_dict({}), bot_names=[bot_name])

    # Check that the job is unaffected
    with instance_context("instance-a"):
        lease = scheduler_a.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
        assert lease is None

        # Attempt to assign leases from the instance which did get the job
        scheduler_a.assign_n_leases_by_priority(capability_hash=hash_from_dict({}), bot_names=[bot_name])

        # This time the job should have a lease
        lease = scheduler_a.synchronize_bot_lease(bot_name, bot_id, BotStatus.OK.value, None)
        assert lease.state == LeaseState.PENDING.value


def test_isolated_instances_list_operations(isolated_controllers):
    controller_a, controller_b = isolated_controllers
    scheduler_a = controller_a.execution_instance.scheduler
    scheduler_b = controller_b.execution_instance.scheduler

    with instance_context("instance-a"):
        mock_queue_job_action(scheduler_a)

    # Check that ListOperations respects instances
    with instance_context("instance-a"):
        operations_a, _ = scheduler_a.list_operations()
    with instance_context("instance-b"):
        operations_b, _ = scheduler_b.list_operations()

    assert len(operations_a) == 1
    assert len(operations_b) == 0


def test_max_queue_size(controller_max_queue_size):
    scheduler = controller_max_queue_size.execution_instance.scheduler

    max_queue_size = scheduler.max_queue_size

    # Queue jobs up to the maximum queue size
    for _ in range(max_queue_size or 10):
        mock_queue_job_action(scheduler, do_not_cache=True, assign=False)

    if max_queue_size is not None:
        # Attempt to queue an additional job, expecting an error
        with pytest.raises(ResourceExhaustedError):
            mock_queue_job_action(scheduler, do_not_cache=True, assign=False)



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
from contextlib import ExitStack
from typing import IO, Callable, Dict, Iterable, Iterator, List, Optional, Tuple, TypeVar

import mmh3

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.decorators import timed
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric
from buildgrid.server.threading import ContextThreadPoolExecutor

from .storage_abc import StorageABC

LOGGER = buildgrid_logger(__name__)


_T = TypeVar("_T")
_R = TypeVar("_R")


# wrapper functions for the bulk StorageABC interfaces
def _bulk_delete_for_storage(storage_digests: Tuple[StorageABC, List[Digest]]) -> List[str]:
    storage, digests = storage_digests
    return storage.bulk_delete(digests)


def _fmb_for_storage(storage_digests: Tuple[StorageABC, List[Digest]]) -> List[Digest]:
    storage, digests = storage_digests
    return storage.missing_blobs(digests)


def _bulk_update_for_storage(
    storage_digests: Tuple[StorageABC, List[Tuple[Digest, bytes]]]
) -> Tuple[StorageABC, List[Status]]:
    storage, digest_tuples = storage_digests
    return storage, storage.bulk_update_blobs(digest_tuples)


def _bulk_read_for_storage(storage_digests: Tuple[StorageABC, List[Digest]]) -> Dict[str, bytes]:
    storage, digests = storage_digests
    return storage.bulk_read_blobs(digests)


class ShardedStorage(StorageABC):
    TYPE = "Sharded"

    def __init__(self, storages: Dict[str, StorageABC], thread_pool_size: Optional[int] = None):
        self._stack = ExitStack()
        if not storages:
            raise ValueError("ShardedStorage requires at least one shard")
        self._storages = storages
        self._threadpool = None
        if thread_pool_size:
            self._threadpool = ContextThreadPoolExecutor(thread_pool_size, "sharded-storage")

    def start(self) -> None:
        if self._threadpool:
            self._stack.enter_context(self._threadpool)
        for storage in self._storages.values():
            self._stack.enter_context(storage)

    def stop(self) -> None:
        self._stack.close()
        LOGGER.info(f"Stopped {type(self).__name__}")

    def _storage_from_digest(self, digest: Digest) -> StorageABC:
        def _score(shard_name: str, digest: Digest) -> int:
            hash = mmh3.hash(f"{shard_name}\t{digest.hash}", signed=False)
            return hash

        shard_name = min(self._storages.keys(), key=lambda name: _score(name, digest))
        return self._storages[shard_name]

    def _partition_digests(self, digests: List[Digest]) -> Dict[StorageABC, List[Digest]]:
        partition: Dict[StorageABC, List[Digest]] = defaultdict(list)
        for digest in digests:
            storage = self._storage_from_digest(digest)
            partition[storage].append(digest)
        return partition

    def _map(self, fn: Callable[[_T], _R], args: Iterable[_T]) -> Iterator[_R]:
        if self._threadpool:
            return self._threadpool.map(fn, args)
        else:
            return map(fn, args)

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        return self._storage_from_digest(digest).has_blob(digest)

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        return self._storage_from_digest(digest).get_blob(digest)

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        self._storage_from_digest(digest).delete_blob(digest)

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        self._storage_from_digest(digest).commit_write(digest, write_session)

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        failed_deletions: List[str] = []
        for result in self._map(_bulk_delete_for_storage, self._partition_digests(digests).items()):
            failed_deletions.extend(result)

        publish_counter_metric(METRIC.STORAGE.DELETE_ERRORS_COUNT, len(failed_deletions), type=self.TYPE)
        return failed_deletions

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        missing_blobs: List[Digest] = []

        for result in self._map(_fmb_for_storage, self._partition_digests(digests).items()):
            missing_blobs.extend(result)

        return missing_blobs

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        partitioned_digests: Dict[StorageABC, List[Tuple[Digest, bytes]]] = defaultdict(list)
        idx_map: Dict[StorageABC, List[int]] = defaultdict(list)
        for orig_idx, digest_tuple in enumerate(blobs):
            storage = self._storage_from_digest(digest_tuple[0])
            partitioned_digests[storage].append(digest_tuple)
            idx_map[storage].append(orig_idx)

        results: List[Status] = [Status(code=code_pb2.INTERNAL, message="inconsistent batch results")] * len(blobs)
        for storage, statuses in self._map(_bulk_update_for_storage, partitioned_digests.items()):
            for status_idx, status in enumerate(statuses):
                results[idx_map[storage][status_idx]] = status
        return results

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        bulk_read_results: Dict[str, bytes] = {}
        for result in self._map(_bulk_read_for_storage, self._partition_digests(digests).items()):
            bulk_read_results.update(result)

        return bulk_read_results



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Cleanup command
=================

Create a BCS cleanup daemon
"""
import signal
import sys
from datetime import timedelta
from os import PathLike
from typing import Any, Dict

import click

from buildgrid.server.app.cli import Context, pass_context, setup_logging
from buildgrid.server.app.settings.config import populate_buildgrid_config
from buildgrid.server.app.settings.parser import load_config
from buildgrid.server.cas.storage.index.index_abc import IndexABC
from buildgrid.server.cleanup.cleanup import CASCleanUp
from buildgrid.server.monitoring import set_monitoring_bus


def parse_size(string: str) -> int:
    """Convert string with suffix representing memory unit, to integer."""
    multipliers = {"K": 1000, "M": 1000000, "G": 1000000000, "T": 1000000000000}
    string = string.upper()
    multiplier = multipliers.get(string[-1], 1)
    amount = float(string[0:-1]) if string[-1] in multipliers else float(string)
    return int(amount * multiplier)


def parse_time(string: str) -> timedelta:
    """Convert string with suffix representing time unit, to timedelta."""
    multipliers = {"M": 60, "H": 3600, "D": 86400, "W": 604800}
    string = string.upper()
    multiplier = multipliers.get(string[-1], 1)
    amount = float(string[0:-1]) if string[-1] in multipliers else float(string)
    return timedelta(seconds=amount * multiplier)


@click.group(name="cleanup", short_help="Start a local cleanup service.")
@pass_context
def cli(context: Context) -> None:
    pass


@cli.command("start", short_help="Setup a new cleanup instance.")
@click.argument("CONFIG", type=click.Path(file_okay=True, dir_okay=False, exists=True, writable=False))
@click.option("-v", "--verbose", count=True, help="Increase log verbosity level.")
@click.option("--dry-run", is_flag=True, help="Do not actually cleanup CAS")
@click.option("--sleep-interval", type=int, help="Seconds to sleep inbetween calls to cleanup")
@click.option(
    "--high-watermark",
    type=str,
    help="Storage size needed to trigger cleanup. K, M, G, and T "
    "suffixes are supported (kilobytes, megabytes, gigabytes, "
    "and terabytes respectively).",
)
@click.option(
    "--low-watermark",
    type=str,
    help="Storage size needed to stop cleanup. K, M, G, and T "
    "suffixes are supported (kilobytes, megabytes, gigabytes, "
    "and terabytes respectively).",
)
@click.option(
    "--batch-size",
    type=str,
    help="Number of bytes to clean up in one go. K, M, G, and T "
    "suffixes are supported (kilobytes, megabytes, gigabytes, "
    "and terabytes respectively).",
)
@click.option(
    "--only-if-unused-for",
    type=str,
    default="0",
    metavar="<INTEGER/FLOAT>[M/H/D/W]",
    help="Number of seconds old a blob must be for cleanup to delete it. "
    "Optional M, H, D, W suffixes are supported (minutes, hours, days, "
    "weeks respectively). For example, if the value 3D is specified. Cleanup will "
    "ignore all blobs which are 3 days or younger. If unspecified, the "
    "value will default to 0 seconds. "
    "Important note: "
    'The value of "--only-if-unused-for" should be greater than that of '
    'the "refresh-accesstime-older-than" parameter found in the sql-index '
    "configuration, which is used to protect the timestamp of blobs "
    "from being updated if they were last accessed too recently. "
    "In other words, to avoid a blob from being cleaned up sooner than expected due "
    "to the effect of suppressed accesstime update, allocate a more generous age "
    "that a blob is allowed to be unused-for.",
)
@pass_context
def start(
    context: Context,
    config: "PathLike[str]",
    verbose: int,
    dry_run: bool,
    high_watermark: str,
    low_watermark: str,
    sleep_interval: int,
    batch_size: str,
    only_if_unused_for: str,
) -> None:
    """Entry point for the bgd-server CLI command group."""
    if dry_run and verbose < 2:
        # If we're doing a dry run, the only thing we care about is the INFO
        # level output, so bump up the verbosity if needed
        verbose = 2
    setup_logging(verbosity=verbose)

    try:
        high_watermark_val = parse_size(high_watermark)
        low_watermark_val = parse_size(low_watermark)
        batch_size_val = parse_size(batch_size)
        only_if_unused_for_val = parse_time(only_if_unused_for)
    except ValueError:
        click.echo(
            "ERROR: Only 'K', 'M', 'G', and 'T' are supported as size suffixes "
            "for the high/low water marks and batch size."
        )
        sys.exit(-1)

    if low_watermark_val > high_watermark_val:
        click.echo("ERROR: The low water mark must be lower than the high water mark.")
        sys.exit(-1)
    batch_size_val = min(batch_size_val, high_watermark_val - low_watermark_val)

    settings = load_config(config)

    try:
        cleanup = _create_cleanup_from_config(
            settings,
            dry_run,
            high_watermark_val,
            low_watermark_val,
            sleep_interval,
            batch_size_val,
            only_if_unused_for_val,
        )
        signal.signal(signal.SIGTERM, cleanup.stop)
        signal.signal(signal.SIGINT, cleanup.stop)
        try:
            cleanup.start()
        finally:
            cleanup.stop()

    except KeyError as e:
        click.echo(f"ERROR: Could not parse config: {e}.\n", err=True)
        sys.exit(-1)

    except KeyboardInterrupt:
        pass

    except Exception as e:
        click.echo(f"ERROR: Uncaught Exception: {e}.\n", err=True)
        sys.exit(-1)


def _create_cleanup_from_config(
    raw_configuration: Dict[str, Any],
    dry_run: bool,
    high_watermark: int,
    low_watermark: int,
    sleep_interval: int,
    batch_size: int,
    only_if_unused_for: timedelta,
) -> CASCleanUp:
    """Parses configuration and setup a fresh server instance."""

    try:
        config = populate_buildgrid_config(raw_configuration)
    except ValueError as e:
        click.echo(f"Error: populating server config: {e}.", err=True)
        sys.exit(-1)

    if monitoring_bus := config.monitoring:
        set_monitoring_bus(monitoring_bus)

    if cleanup := config.cleanup:
        instance_indexes = {}
        for cleanup_entry in cleanup:
            instance_indexes[cleanup_entry.name] = cleanup_entry.index

        return CASCleanUp(
            indexes=instance_indexes,
            dry_run=dry_run,
            high_watermark=high_watermark,
            low_watermark=low_watermark,
            sleep_interval=sleep_interval,
            batch_size=batch_size,
            only_if_unused_for=only_if_unused_for,
            monitor=config.monitoring is not None,
        )

    try:
        instance_indexes = {}
        for instance in config.instances:
            for storage in instance.storages:
                if isinstance(storage, IndexABC):
                    instance_indexes[instance.name] = storage
                    break
            else:
                click.echo(f"Warning: Skipping instance {instance.name}.", err=False)

    except KeyError as e:
        click.echo(f"Error: Storage/Index missing from configuration: {e}.", err=True)
        sys.exit(-1)

    if instance_indexes:
        return CASCleanUp(
            indexes=instance_indexes,
            dry_run=dry_run,
            high_watermark=high_watermark,
            low_watermark=low_watermark,
            sleep_interval=sleep_interval,
            batch_size=batch_size,
            only_if_unused_for=only_if_unused_for,
            monitor=config.monitoring is not None,
        )

    click.echo("Error: Did not find any indexes to clean.", err=True)
    sys.exit(-1)



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
S3 Action Cache
==================

Implements an Action Cache using S3 to store cache entries.

"""

import io
import logging
from typing import Any, Dict, Optional

import boto3
from botocore.config import Config as BotoConfig
from botocore.exceptions import ClientError

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.cas.storage.storage_abc import StorageABC
from buildgrid.server.enums import ActionCacheEntryType
from buildgrid.server.exceptions import NotFoundError, StorageFullError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.s3 import s3utils

LOGGER = buildgrid_logger(__name__)


class S3ActionCache(ActionCacheABC):
    def __init__(
        self,
        storage: StorageABC,
        allow_updates: bool = True,
        cache_failed_actions: bool = True,
        entry_type: Optional[ActionCacheEntryType] = ActionCacheEntryType.ACTION_RESULT_DIGEST,
        migrate_entries: Optional[bool] = False,
        bucket: Optional[str] = None,
        endpoint_url: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        config: Optional[BotoConfig] = BotoConfig(),
    ):
        """Initialises a new ActionCache instance using S3 to persist the action cache.

        Args:
            storage (StorageABC): storage backend instance to be used to store ActionResults.
            allow_updates (bool): allow the client to write to storage
            cache_failed_actions (bool): whether to store failed actions in the Action Cache
            entry_type (ActionCacheEntryType): whether to store ActionResults or their digests.
            migrate_entries (bool): if set, migrate entries that contain a value with
                a different `EntryType` to `entry_type` as they are accessed
                (False by default).

            bucket (str): Name of bucket
            endpoint_url (str): URL of endpoint.
            aws_access_key_id (str): S3-ACCESS-KEY
            aws_secret_access_key (str): S3-SECRET-KEY
        """
        super().__init__(storage=storage, allow_updates=allow_updates)

        self._entry_type = entry_type
        self._migrate_entries = migrate_entries

        self._cache_failed_actions = cache_failed_actions
        assert bucket is not None
        self._bucket_template = bucket

        # Boto logs can be very verbose, restrict to WARNING
        for boto_logger_name in ["boto3", "botocore", "s3transfer", "urllib3"]:
            boto_logger = logging.getLogger(boto_logger_name)
            boto_logger.setLevel(max(boto_logger.level, logging.WARNING))

        # Only pass arguments with a value to support testing with moto_server
        client_kwargs: Dict[str, Any] = {}
        if endpoint_url is not None:
            client_kwargs["endpoint_url"] = endpoint_url
        if aws_access_key_id is not None:
            client_kwargs["aws_access_key_id"] = aws_access_key_id
        if aws_secret_access_key is not None:
            client_kwargs["aws_secret_access_key"] = aws_secret_access_key
        client_kwargs["config"] = config

        self._s3cache = boto3.client("s3", **client_kwargs)

    # --- Public API ---
    @property
    def allow_updates(self) -> bool:
        return self._allow_updates

    def get_action_result(self, action_digest: Digest) -> ActionResult:
        """Retrieves the cached ActionResult for the given Action digest.

        Args:
            action_digest: The digest to get the result for

        Returns:
            The cached ActionResult matching the given key or raises
            NotFoundError.
        """
        action_result = self._get_action_result(action_digest)
        if action_result is not None:
            if self.referenced_blobs_still_exist(action_digest, action_result):
                return action_result

            if self._allow_updates:
                LOGGER.debug(
                    "Removing action from cache due to missing blobs in CAS.",
                    tags=dict(digest=action_digest),
                )
                self._delete_key_from_cache(action_digest)

        raise NotFoundError(f"Key not found: {action_digest.hash}/{action_digest.size_bytes}")

    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        """Stores the result in cache for the given key.

        Args:
            action_digest (Digest): digest of Action to update
            action_result (ActionResult): ActionResult to store.
        """
        if not self._allow_updates:
            raise NotImplementedError("Updating cache not allowed")

        if self._cache_failed_actions or action_result.exit_code == 0:
            assert self._storage, "Storage used before initialization"
            action_result_digest = self._storage.put_message(action_result)

            if self._entry_type == ActionCacheEntryType.ACTION_RESULT_DIGEST:
                self._update_cache_key(action_digest, action_result_digest.SerializeToString())
            else:
                self._update_cache_key(action_digest, action_result.SerializeToString())

            LOGGER.info("Result cached for action.", tags=dict(digest=action_digest))

    # --- Private API ---
    def _get_action_result(self, digest: Digest) -> Optional[ActionResult]:
        """Get an `ActionResult` from the cache.

        If present, returns the `ActionResult` corresponding to the given digest.
        Otherwise returns None.

        When configured to do so, if the value stored in the entry in S3 contains
        a different type, it will convert it to `entry_type`.

        Args:
            digest: Action digest to get the associated ActionResult digest for

        Returns:
            The `ActionResult` or None if the digest is not present
        """
        value_in_cache = self._get_value_from_cache(digest)
        if not value_in_cache:
            return None

        # Attempting to parse the entry as a `Digest` first:
        action_result_digest = Digest.FromString(value_in_cache)
        if len(action_result_digest.hash) == len(digest.hash):
            # The cache contains the `Digest` of the `ActionResult`:
            assert self._storage, "Storage used before initialization"
            action_result = self._storage.get_message(action_result_digest, ActionResult)

            # If configured, update the entry to contain an `ActionResult`:
            if self._entry_type == ActionCacheEntryType.ACTION_RESULT and self._migrate_entries:
                LOGGER.debug("Converting entry from Digest to ActionResult.", tags=dict(digest=digest))
                assert action_result, "Returned result was none"
                self._update_cache_key(digest, action_result.SerializeToString())

        else:
            action_result = ActionResult.FromString(value_in_cache)

            # If configured, update the entry to contain a `Digest`:
            if self._entry_type == ActionCacheEntryType.ACTION_RESULT_DIGEST and self._migrate_entries:
                LOGGER.debug("Converting entry from ActionResult to Digest.", tags=dict(digest=digest))
                assert self._storage, "Storage used before initialization"
                action_result_digest = self._storage.put_message(action_result)
                self._update_cache_key(digest, action_result_digest.SerializeToString())

        return action_result

    def _get_value_from_cache(self, digest: Digest) -> Optional[bytes]:
        """Get the bytes stored in cache for the given Digest.

        Args:
            digest: Action digest to get the associated bytes in S3

        Returns:
            bytes or None if the digest is not present
        """

        try:
            s3object = self._get_s3object(digest)
            s3object.fileobj = io.BytesIO()
            s3utils.get_object(self._s3cache, s3object)
            s3object.fileobj.seek(0)
            return s3object.fileobj.read()
        except ClientError as e:
            if e.response["Error"]["Code"] not in ["404", "NoSuchKey"]:
                raise
            return None

    def _update_cache_key(self, digest: Digest, value: bytes) -> None:
        try:
            s3object = self._get_s3object(digest)
            s3object.fileobj = io.BytesIO(value)
            s3object.filesize = len(value)
            s3utils.put_object(self._s3cache, s3object)
        except ClientError as error:
            if error.response["Error"]["Code"] == "QuotaExceededException":
                raise StorageFullError("ActionCache S3 Quota Exceeded.") from error
            raise error

    def _delete_key_from_cache(self, digest: Digest) -> None:
        """Remove an entry from the ActionCache

        Args:
            digest: entry to remove from the ActionCache

        Returns:
            None
        """
        if not self._allow_updates:
            raise NotImplementedError("Updating cache not allowed")

        self._s3cache.delete_object(Bucket=self._get_bucket_name(digest), Key=self._get_key(digest))

    @staticmethod
    def _get_key(digest: Digest) -> str:
        """
        Given a `Digest`, returns the key used to store its
        corresponding entry in S3.
        """
        return f"{digest.hash}_{digest.size_bytes}"

    def _get_bucket_name(self, digest: Digest) -> str:
        """Return the formatted bucket name for a given digest.

        This formats the bucket template defined in the configuration file
        using a digest to be stored in the cache. This allows the cache
        contents to be sharded across multiple S3 buckets, allowing us to
        cache more Actions than can be stored in a single S3 bucket.

        Currently the only variable interpolated into the template is
        ``digest``, which contains the hash part of the Digest. A template
        string which includes undefined variables will result in a
        non-functional ActionCache.

        Args:
            digest (Digest): The digest to get the bucket name for.

        Returns:
            str: The bucket name corresponding to the given Digest.

        """
        try:
            return self._bucket_template.format(digest=digest.hash)
        except IndexError:
            LOGGER.error(
                (
                    "Could not generate bucket name for digest. "
                    "This is either a misconfiguration in the BuildGrid S3 "
                    "ActionCache bucket configuration, or a badly formed request."
                ),
                tags=dict(digest=digest),
            )
            raise

    def _get_s3object(self, digest: Digest) -> s3utils.S3Object:
        return s3utils.S3Object(self._get_bucket_name(digest), self._get_key(digest))



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import random
import threading
import time
from typing import IO, TYPE_CHECKING, Any, BinaryIO, Callable, Dict, List, Mapping, Optional, Sequence, Union, cast

import botocore
import pycurl

from buildgrid.server.settings import (
    S3_MAX_RETRIES,
    S3_MULTIPART_MAX_CONCURRENT_PARTS,
    S3_MULTIPART_PART_SIZE,
    S3_TIMEOUT_CONNECT,
    S3_TIMEOUT_READ,
    S3_USERAGENT_NAME,
)

# https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html
_RETRIABLE_HTTP_STATUS_CODES = (408, 429, 500, 502, 503, 504, 509)
_RETRIABLE_S3_ERROR_CODES = (
    "Throttling",
    "ThrottlingException",
    "ThrottledException",
    "RequestThrottledException",
    "ProvisionedThroughputExceededException",
)
# Maximum backoff in seconds
_MAX_BACKOFF = 20
# Maximum requests to run in parallel via CurlMulti
_MAX_CURLMULTI_CONNECTIONS = 10

if TYPE_CHECKING:
    from mypy_boto3_s3 import Client as S3Client


class _CurlLocal(threading.local):
    def __init__(self) -> None:
        self.curlmulti = pycurl.CurlMulti()
        self.curlmulti.setopt(pycurl.M_MAX_TOTAL_CONNECTIONS, _MAX_CURLMULTI_CONNECTIONS)


_curlLocal = _CurlLocal()


class S3Object:
    def __init__(self, bucket: str, key: str) -> None:
        self.bucket = bucket
        self.key = key
        self.fileobj: Optional[IO[bytes]] = None
        self.filesize: Optional[int] = None
        self.error: Optional[Exception] = None
        self.status_code: Optional[int] = None
        self._method: str = ""
        self._errfileobj: Optional[IO[bytes]] = None
        self._response_headers: Dict[str, str] = {}

    @property
    def response_headers(self) -> Dict[str, str]:
        return self._response_headers

    # Function to process HTTP response headers
    def _header_function(self, header_line: bytes) -> None:
        header = header_line.decode("ascii")

        # Skip status line
        if ":" not in header:
            return

        name, value = header.split(":", maxsplit=1)
        name = name.strip().lower()
        value = value.strip()

        self._response_headers[name] = value


class UploadPart(io.BufferedIOBase):
    def __init__(self, upload_id: str, number: int, file: IO[bytes], eof: int, size: int, offset: int) -> None:
        super().__init__()
        self._upload_id = upload_id
        self._number = number
        self._response = io.BytesIO()
        self._file = file
        self._content = None
        try:
            self._fd: Optional[int] = file.fileno()
        except OSError:
            # The "file" doesn't have a file descriptor, its probably a BytesIO.
            # Read our part now so that we don't need to cope with thread safety
            # when `UploadPart.read` is called.
            self._fd = None
            old_position = file.tell()
            file.seek(offset)
            self._content = file.read(size)
            file.seek(old_position)

        self._size = size
        self._start = offset
        self._end = min(eof, offset + size)
        self._read_offset = 0

    @property
    def upload_id(self) -> str:
        return self._upload_id

    @property
    def number(self) -> int:
        return self._number

    @property
    def response(self) -> BinaryIO:
        return self._response

    def __len__(self) -> int:
        return self._end - self._start

    def readable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return False

    def writable(self) -> bool:
        return True

    def read(self, size: Optional[int] = -1) -> bytes:
        # If we have a real file underlying this part, then we want to do an
        # `os.pread` for just the part that is relevant.
        if self._fd is not None:
            if size is None or size == -1:
                size = self._size

            # Calculate the actual read offset and make sure we're within our
            # section of the file.
            offset = self._start + self._read_offset
            if offset >= self._end:
                return b""

            # Make sure we only read up to the end of our section of the file,
            # in case the size requested is larger than the number of bytes
            # remaining in our section
            size = min(size, self._end - offset)
            content = os.pread(self._fd, size, offset)
            self._read_offset += size
            return content

        # Otherwise we can just return our pre-determined slice of the actual
        # contents. This case should only be reached when MAX_IN_MEMORY_BLOB_SIZE_BYTES
        # is the same as or larger than S3_MAX_UPLOAD_SIZE, which should ideally
        # never be the case.
        else:
            if self._content is None:
                raise ValueError(
                    f"Part {self._number} of upload {self._upload_id} is backed "
                    "by a BytesIO but the content couldn't be read when the part "
                    "was instantiated."
                )
            return self._content

    def write(self, b: bytes) -> int:  # type: ignore[override]
        return self._response.write(b)


def _curl_handle_for_s3(
    s3: "S3Client", method: str, s3object: S3Object, extra_params: Optional[Mapping[str, Union[str, int]]] = None
) -> pycurl.Curl:
    if extra_params is None:
        extra_params = {}
    s3object._method = method
    params: Dict[str, Union[str, int]] = {"Bucket": s3object.bucket, "Key": s3object.key, **extra_params}
    url = s3.generate_presigned_url(method, Params=params, ExpiresIn=3600)
    c = pycurl.Curl()
    c.s3object = s3object  # type: ignore
    c.setopt(pycurl.USERAGENT, S3_USERAGENT_NAME)
    c.setopt(pycurl.CONNECTTIMEOUT, S3_TIMEOUT_CONNECT)
    c.setopt(pycurl.TIMEOUT, S3_TIMEOUT_READ)
    c.setopt(pycurl.FAILONERROR, True)
    c.setopt(pycurl.URL, url)
    c.setopt(pycurl.HEADERFUNCTION, s3object._header_function)
    return c


# TODO Don't put random attributes on curl like this??
def c_s3ojb(c: pycurl.Curl) -> S3Object:
    return c.s3object  # type: ignore


def _curl_should_retry(c: pycurl.Curl, errno: int) -> bool:
    if errno in (
        pycurl.E_COULDNT_CONNECT,
        pycurl.E_SEND_ERROR,
        pycurl.E_RECV_ERROR,
        pycurl.E_OPERATION_TIMEDOUT,
        pycurl.E_PARTIAL_FILE,
    ):
        # Retry on network and timeout errors
        return True

    if errno == pycurl.E_HTTP_RETURNED_ERROR:
        s3obj = c_s3ojb(c)
        if s3obj.status_code in _RETRIABLE_HTTP_STATUS_CODES:
            # Retry on 'Request Timeout', 'Too Many Requests' and transient server errors
            return True

        if error_response := getattr(s3obj.error, "response", None):
            if error_response["Error"]["Code"] in _RETRIABLE_S3_ERROR_CODES:
                return True

    return False


def _curl_multi_run(
    objects: Sequence[S3Object], curl_handle_func: Callable[[S3Object], pycurl.Curl], attempt: int = 1
) -> None:
    m = _curlLocal.curlmulti
    for s3object in objects:
        c = curl_handle_func(s3object)
        m.add_handle(c)

    while True:
        ret, active_handles = m.perform()
        if ret == pycurl.E_CALL_MULTI_PERFORM:
            # More processing required
            continue

        if active_handles:
            # Wait for next event
            m.select(15.0)
        else:
            # All operations complete
            break

    num_q, ok_list, err_list = m.info_read()
    assert num_q == 0

    retry_objects = []
    for c in ok_list:
        s3obj = c_s3ojb(c)
        s3obj.status_code = c.getinfo(pycurl.HTTP_CODE)  # type: ignore[no-untyped-call]
        m.remove_handle(c)
        c.close()
    for c, errno, errmsg in err_list:
        s3obj = c_s3ojb(c)
        if errno == pycurl.E_HTTP_RETURNED_ERROR:
            s3obj.status_code = c.getinfo(pycurl.HTTP_CODE)  # type: ignore[no-untyped-call]
            response: Dict[str, Any] = {}
            response["status_code"] = s3obj.status_code
            response["headers"] = s3obj._response_headers
            if (errfileobj := s3obj._errfileobj) is None:
                response["body"] = b""
            else:
                errfileobj.seek(0)
                response["body"] = errfileobj.read()
                errfileobj.truncate(0)
            parser = botocore.parsers.RestXMLParser()
            # TODO: botocore safely handles `None` being passed here, but it is
            # probably best to rework this to get the correct `Shape` to match
            # the type hints from boto3-stubs
            parsed_response = parser.parse(response, None)  # type: ignore[arg-type]
            s3obj.error = botocore.exceptions.ClientError(parsed_response, s3obj._method)
        else:
            s3obj.error = pycurl.error(errmsg)

        if attempt < S3_MAX_RETRIES and _curl_should_retry(c, errno):
            s3obj.status_code = None
            s3obj.error = None
            retry_objects.append(s3obj)

        m.remove_handle(c)
        c.close()

    if retry_objects and attempt < S3_MAX_RETRIES:
        # Wait between attempts with truncated exponential backoff with jitter
        exp_backoff = 2 ** (attempt - 1)
        exp_backoff_with_jitter = random.random() * exp_backoff
        time.sleep(min(exp_backoff_with_jitter, _MAX_BACKOFF))

        _curl_multi_run(retry_objects, curl_handle_func, attempt=attempt + 1)


def head_objects(s3: "S3Client", objects: Sequence[S3Object]) -> None:
    def curl_handle_func(s3object: S3Object) -> pycurl.Curl:
        c = _curl_handle_for_s3(s3, "head_object", s3object)
        c.setopt(pycurl.NOBODY, True)
        return c

    _curl_multi_run(objects, curl_handle_func)


def head_object(s3: "S3Client", s3object: S3Object) -> None:
    head_objects(s3, [s3object])
    if s3object.error is not None:
        raise s3object.error


def set_s3_timeout(
    c: pycurl.Curl,
    s3object: S3Object,
    timeout_seconds_per_kilobyte: Optional[float],
    timeout_min_seconds: float,
) -> None:
    timeout = timeout_min_seconds
    if s3object.filesize is not None and timeout_seconds_per_kilobyte is not None:
        timeout = max(timeout, s3object.filesize * timeout_seconds_per_kilobyte / 1024)
    c.setopt(pycurl.TIMEOUT, int(timeout))


def get_objects(
    s3: "S3Client",
    objects: Sequence[S3Object],
    timeout_seconds_per_kilobyte: Optional[float] = None,
    timeout_min_seconds: float = S3_TIMEOUT_READ,
) -> None:
    def curl_handle_func(s3object: S3Object) -> pycurl.Curl:
        c = _curl_handle_for_s3(s3, "get_object", s3object)
        c.setopt(pycurl.WRITEDATA, s3object.fileobj)
        set_s3_timeout(c, s3object, timeout_seconds_per_kilobyte, timeout_min_seconds)
        s3object._errfileobj = s3object.fileobj
        return c

    _curl_multi_run(objects, curl_handle_func)


def get_object(
    s3: "S3Client",
    s3object: S3Object,
    timeout_seconds_per_kilobyte: Optional[float] = None,
    timeout_min_seconds: float = S3_TIMEOUT_READ,
) -> None:
    get_objects(s3, [s3object], timeout_seconds_per_kilobyte, timeout_min_seconds)
    if s3object.error is not None:
        raise s3object.error


def put_objects(
    s3: "S3Client",
    objects: Sequence[S3Object],
    timeout_seconds_per_kilobyte: Optional[float] = None,
    timeout_min_seconds: float = S3_TIMEOUT_READ,
) -> None:
    def curl_handle_func(s3object: S3Object) -> pycurl.Curl:
        c = _curl_handle_for_s3(s3, "put_object", s3object)
        c.setopt(pycurl.READDATA, s3object.fileobj)
        c.setopt(pycurl.INFILESIZE_LARGE, s3object.filesize)
        c.setopt(pycurl.UPLOAD, 1)
        set_s3_timeout(c, s3object, timeout_seconds_per_kilobyte, timeout_min_seconds)
        s3object._errfileobj = io.BytesIO()
        c.setopt(pycurl.WRITEDATA, s3object._errfileobj)
        return c

    _curl_multi_run(objects, curl_handle_func)


def put_object(
    s3: "S3Client",
    s3object: S3Object,
    timeout_seconds_per_kilobyte: Optional[float] = None,
    timeout_min_seconds: float = S3_TIMEOUT_READ,
) -> None:
    put_objects(s3, [s3object], timeout_seconds_per_kilobyte, timeout_min_seconds)
    if s3object.error is not None:
        raise s3object.error


ParsedResponse = Dict[str, Union[str, Dict[str, Union[str, int, Dict[str, str]]]]]


def _parse_s3_response(s3: "S3Client", response: BinaryIO, s3object: S3Object, shape_name: str) -> ParsedResponse:
    response_dict: Dict[str, Union[Optional[int], Dict[str, str], bytes]] = {}
    response_dict["status_code"] = s3object.status_code
    response_dict["headers"] = s3object.response_headers
    response.seek(0)
    response_dict["body"] = response.read()

    parser = botocore.parsers.RestXMLParser()
    shape = s3.meta.service_model.shape_for(shape_name)
    return cast(ParsedResponse, parser.parse(response_dict, shape))


def start_multipart_upload(s3: "S3Client", s3object: S3Object) -> str:
    response = io.BytesIO()

    def curl_handle_func(_s3object: S3Object) -> pycurl.Curl:
        c = _curl_handle_for_s3(s3, "create_multipart_upload", _s3object)
        c.setopt(pycurl.WRITEDATA, response)
        c.setopt(pycurl.POST, 1)
        return c

    _curl_multi_run([s3object], curl_handle_func)
    if s3object.error is not None:
        raise s3object.error

    parsed_response = _parse_s3_response(s3, response, s3object, "CreateMultipartUploadOutput")
    return parsed_response["UploadId"]  # type: ignore


def upload_parts(s3: "S3Client", s3object: S3Object, parts: Sequence[UploadPart]) -> Dict[int, str]:
    s3object_map = {S3Object(s3object.bucket, s3object.key): part for part in parts}

    def curl_handle_func(_s3object: S3Object) -> pycurl.Curl:
        part = s3object_map[_s3object]
        params: Dict[str, Union[int, str]] = {"UploadId": part.upload_id, "PartNumber": part.number}
        c = _curl_handle_for_s3(s3, "upload_part", _s3object, extra_params=params)
        c.setopt(pycurl.READDATA, part)
        c.setopt(pycurl.UPLOAD, 1)
        c.setopt(pycurl.INFILESIZE_LARGE, len(part))
        c.setopt(pycurl.WRITEDATA, part)
        return c

    _curl_multi_run(list(s3object_map.keys()), curl_handle_func)

    errors: List[Exception] = [_s3object.error for _s3object in s3object_map.keys() if _s3object.error is not None]
    if len(errors) > 0:
        raise errors[0]

    uploaded: Dict[int, str] = {}
    for _s3object, part in s3object_map.items():
        response = _parse_s3_response(s3, part.response, _s3object, "UploadPartOutput")
        uploaded[part.number] = response["ResponseMetadata"]["HTTPHeaders"]["etag"]  # type: ignore
    return uploaded


def complete_multipart_upload(
    s3: "S3Client", s3object: S3Object, upload_id: str, parts: Mapping[int, str]
) -> ParsedResponse:
    # Use the boto3 client directly here, rather than a presigned URL. This is
    # necessary because boto3's URL presigning is broken for `complete_multipart_upload`
    # when using s3v4 auth.
    #
    # See https://github.com/boto/boto3/issues/2192
    return cast(
        ParsedResponse,
        s3.complete_multipart_upload(
            Bucket=s3object.bucket,
            Key=s3object.key,
            UploadId=upload_id,
            MultipartUpload={"Parts": [{"ETag": tag, "PartNumber": number} for number, tag in parts.items()]},
        ),
    )


def _list_multipart_parts(s3: "S3Client", s3object: S3Object, upload_id: str) -> ParsedResponse:
    response = io.BytesIO()

    def curl_handle_func(_s3object: S3Object) -> pycurl.Curl:
        params = {"UploadId": upload_id}
        c = _curl_handle_for_s3(s3, "list_parts", _s3object, extra_params=params)
        c.setopt(pycurl.WRITEDATA, response)
        return c

    _curl_multi_run([s3object], curl_handle_func)
    if s3object.error is not None:
        raise s3object.error
    return _parse_s3_response(s3, response, s3object, "ListPartsOutput")


def abort_multipart_upload(s3: "S3Client", s3object: S3Object, upload_id: str) -> None:
    def curl_handle_func(_s3object: S3Object) -> pycurl.Curl:
        params = {"UploadId": upload_id}
        c = _curl_handle_for_s3(s3, "abort_multipart_upload", _s3object, extra_params=params)
        c.setopt(pycurl.CUSTOMREQUEST, "DELETE")
        return c

    parts = _list_multipart_parts(s3, s3object, upload_id)

    # We need to iterate here in case any part uploads slip through in a race
    # against the AbortMultipartUpload call.
    #
    # See https://docs.aws.amazon.com/AmazonS3/latest/API/API_AbortMultipartUpload.html
    while len(parts.get("Parts", [])) > 0:
        _curl_multi_run([s3object], curl_handle_func)
        try:
            parts = _list_multipart_parts(s3, s3object, upload_id)
        except botocore.exceptions.ClientError as e:
            code = e.response.get("Error", {}).get("Code")
            if code == "404":
                # 404 error here means that the multipart upload is properly aborted.
                break
            raise e


def multipart_upload(s3: "S3Client", s3object: S3Object) -> None:
    if s3object.fileobj is None or s3object.filesize is None:
        raise TypeError("S3Object provided to multipart upload didn't contain a file.")

    upload_id = start_multipart_upload(s3, s3object)

    try:
        part_number = 1
        parts: Dict[int, str] = {}
        queue: List[UploadPart] = []
        while (part_number - 1) * S3_MULTIPART_PART_SIZE < s3object.filesize:
            part = UploadPart(
                upload_id=upload_id,
                number=part_number,
                file=s3object.fileobj,
                eof=s3object.filesize,
                size=S3_MULTIPART_PART_SIZE,
                offset=(part_number - 1) * S3_MULTIPART_PART_SIZE,
            )
            queue.append(part)

            part_number += 1

            if len(queue) >= S3_MULTIPART_MAX_CONCURRENT_PARTS:
                uploaded = upload_parts(s3, s3object, queue)
                parts.update(uploaded)
                queue = []
        uploaded = upload_parts(s3, s3object, queue)
        parts.update(uploaded)

        complete_multipart_upload(s3, s3object, upload_id, parts)
    except Exception as e:
        abort_multipart_upload(s3, s3object, upload_id)
        raise e



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright (C) 2018 Bloomberg LP

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 <http://www.apache.org/licenses/LICENSE-2.0>

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.duration_pb2
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class BusMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SEQUENCE_NUMBER_FIELD_NUMBER: builtins.int
    LOG_RECORD_FIELD_NUMBER: builtins.int
    METRIC_RECORD_FIELD_NUMBER: builtins.int
    sequence_number: builtins.int
    """The position of this message in the bus stream."""
    @property
    def log_record(self) -> global___LogRecord: ...
    @property
    def metric_record(self) -> global___MetricRecord: ...
    def __init__(
        self,
        *,
        sequence_number: builtins.int = ...,
        log_record: global___LogRecord | None = ...,
        metric_record: global___MetricRecord | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["log_record", b"log_record", "metric_record", b"metric_record", "record", b"record"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["log_record", b"log_record", "metric_record", b"metric_record", "record", b"record", "sequence_number", b"sequence_number"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["record", b"record"]) -> typing_extensions.Literal["log_record", "metric_record"] | None: ...

global___BusMessage = BusMessage

@typing_extensions.final
class LogRecord(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Level:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _LevelEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[LogRecord._Level.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        NOTSET: LogRecord._Level.ValueType  # 0
        DEBUG: LogRecord._Level.ValueType  # 1
        """Debug message severity level."""
        INFO: LogRecord._Level.ValueType  # 2
        """Information message severity level."""
        WARNING: LogRecord._Level.ValueType  # 3
        """Warning message severity level."""
        ERROR: LogRecord._Level.ValueType  # 4
        """Error message severity level."""
        CRITICAL: LogRecord._Level.ValueType  # 5
        """Critical message severity level."""

    class Level(_Level, metaclass=_LevelEnumTypeWrapper): ...
    NOTSET: LogRecord.Level.ValueType  # 0
    DEBUG: LogRecord.Level.ValueType  # 1
    """Debug message severity level."""
    INFO: LogRecord.Level.ValueType  # 2
    """Information message severity level."""
    WARNING: LogRecord.Level.ValueType  # 3
    """Warning message severity level."""
    ERROR: LogRecord.Level.ValueType  # 4
    """Error message severity level."""
    CRITICAL: LogRecord.Level.ValueType  # 5
    """Critical message severity level."""

    @typing_extensions.final
    class MetadataEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    CREATION_TIMESTAMP_FIELD_NUMBER: builtins.int
    DOMAIN_FIELD_NUMBER: builtins.int
    LEVEL_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    @property
    def creation_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """When the record has been created."""
    domain: builtins.str
    """The domain name for the record."""
    level: global___LogRecord.Level.ValueType
    """The severity level of the record."""
    message: builtins.str
    """The human-readable record's message."""
    @property
    def metadata(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """An optional list of additional metadata."""
    def __init__(
        self,
        *,
        creation_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        domain: builtins.str = ...,
        level: global___LogRecord.Level.ValueType = ...,
        message: builtins.str = ...,
        metadata: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["creation_timestamp", b"creation_timestamp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["creation_timestamp", b"creation_timestamp", "domain", b"domain", "level", b"level", "message", b"message", "metadata", b"metadata"]) -> None: ...

global___LogRecord = LogRecord

@typing_extensions.final
class MetricRecord(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[MetricRecord._Type.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        NONE: MetricRecord._Type.ValueType  # 0
        COUNTER: MetricRecord._Type.ValueType  # 1
        """A metric for counting."""
        TIMER: MetricRecord._Type.ValueType  # 2
        """A metric for measuring a duration."""
        GAUGE: MetricRecord._Type.ValueType  # 3
        """A metric in arbitrary value."""
        DISTRIBUTION: MetricRecord._Type.ValueType  # 4
        """A metric distribution semantics"""

    class Type(_Type, metaclass=_TypeEnumTypeWrapper): ...
    NONE: MetricRecord.Type.ValueType  # 0
    COUNTER: MetricRecord.Type.ValueType  # 1
    """A metric for counting."""
    TIMER: MetricRecord.Type.ValueType  # 2
    """A metric for measuring a duration."""
    GAUGE: MetricRecord.Type.ValueType  # 3
    """A metric in arbitrary value."""
    DISTRIBUTION: MetricRecord.Type.ValueType  # 4
    """A metric distribution semantics"""

    @typing_extensions.final
    class MetadataEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    CREATION_TIMESTAMP_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    COUNT_FIELD_NUMBER: builtins.int
    DURATION_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    @property
    def creation_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """When the metric has been created."""
    type: global___MetricRecord.Type.ValueType
    """The type of metric, see Type."""
    name: builtins.str
    """The name identifying the metric."""
    count: builtins.float
    """Set for Type.COUNTER and Type.DISTRIBUTION metrics."""
    @property
    def duration(self) -> google.protobuf.duration_pb2.Duration:
        """Set for Type.TIMER metrics."""
    value: builtins.float
    """Set for Type.GAUGE metrics."""
    @property
    def metadata(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """An optional list of additional metadata."""
    def __init__(
        self,
        *,
        creation_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        type: global___MetricRecord.Type.ValueType = ...,
        name: builtins.str = ...,
        count: builtins.float = ...,
        duration: google.protobuf.duration_pb2.Duration | None = ...,
        value: builtins.float = ...,
        metadata: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["count", b"count", "creation_timestamp", b"creation_timestamp", "data", b"data", "duration", b"duration", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["count", b"count", "creation_timestamp", b"creation_timestamp", "data", b"data", "duration", b"duration", "metadata", b"metadata", "name", b"name", "type", b"type", "value", b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["data", b"data"]) -> typing_extensions.Literal["count", "duration", "value"] | None: ...

global___MetricRecord = MetricRecord



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
StorageABC
==================

The abstract base class for storage providers.
"""

import abc
import io
from tempfile import TemporaryFile
from typing import IO, Any, Dict, Iterator, List, Optional, Tuple, Type, TypeVar

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest, Directory
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.exceptions import NotFoundError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import timer
from buildgrid.server.settings import HASH, MAX_IN_MEMORY_BLOB_SIZE_BYTES
from buildgrid.server.types import MessageType

LOGGER = buildgrid_logger(__name__)

M = TypeVar("M", bound=MessageType)


def create_write_session(digest: Digest) -> IO[bytes]:
    """
    Return a file-like object to which a blob's contents could be written.

    For large files, to avoid excess memory usage, upload to temporary file.
    For small files we can work in memory for performance.
    """

    if digest.size_bytes > MAX_IN_MEMORY_BLOB_SIZE_BYTES:
        return TemporaryFile()
    return io.BytesIO()


T = TypeVar("T", bound="StorageABC")


class StorageABC(abc.ABC):
    TYPE: str

    def __enter__(self: T) -> T:
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()

    def start(self) -> None:
        pass

    def stop(self) -> None:
        LOGGER.info(f"Stopped {type(self).__name__}")

    @abc.abstractmethod
    def has_blob(self, digest: Digest) -> bool:
        """Return True if the blob with the given instance/digest exists."""

    @abc.abstractmethod
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        """Return a file-like object containing the blob. Most implementations
        will read the entire file into memory and return a `BytesIO` object.
        Eventually this should be corrected to handle files which cannot fit
        into memory.

        The file-like object must be readable and seekable.

        If the blob isn't present in storage, return None.
        """

    @abc.abstractmethod
    def delete_blob(self, digest: Digest) -> None:
        """Delete the blob from storage if it's present."""

    @abc.abstractmethod
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        """Store the contents for a digest.

        The storage object is not responsible for verifying that the data
        written to the write_session actually matches the digest. The caller
        must do that.
        """

    @abc.abstractmethod
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        """Delete a list of blobs from storage."""

    @abc.abstractmethod
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        """Return a container containing the blobs not present in CAS."""

    @abc.abstractmethod
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        """Given a container of (digest, value) tuples, add all the blobs
        to CAS. Return a list of Status objects corresponding to the
        result of uploading each of the blobs.

        Unlike in `commit_write`, the storage object will verify that each of
        the digests matches the provided data.
        """

    @abc.abstractmethod
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        """Given an iterable container of digests, return a
        {hash: file-like object} dictionary corresponding to the blobs
        represented by the input digests.

        Each file-like object must be readable and seekable.
        """

    def put_message(self, message: MessageType) -> Digest:
        """Store the given Protobuf message in CAS, returning its digest."""
        message_blob = message.SerializeToString()
        digest = Digest(hash=HASH(message_blob).hexdigest(), size_bytes=len(message_blob))
        with create_write_session(digest) as session:
            session.write(message_blob)
            self.commit_write(digest, session)
        return digest

    def get_message(self, digest: Digest, message_type: Type[M]) -> Optional[M]:
        """Retrieve the Protobuf message with the given digest and type from
        CAS. If the blob is not present, returns None.
        """
        message_blob = self.get_blob(digest)
        if message_blob is None:
            return None
        try:
            return message_type.FromString(message_blob.read())
        finally:
            message_blob.close()

    def get_tree(self, root_digest: Digest, raise_on_missing_subdir: bool = False) -> Iterator[Directory]:
        # From the spec, a NotFound response only occurs if the root directory is missing.
        with timer(METRIC.STORAGE.GET_TREE_DURATION, type=self.TYPE):
            root_directory = self.get_message(root_digest, Directory)
            if root_directory is None:
                raise NotFoundError(f"Root digest not found: {root_digest.hash}/{root_digest.size_bytes}")
            yield root_directory

            queue = [subdir.digest for subdir in root_directory.directories]
            while queue:
                blobs = self.bulk_read_blobs(queue)

                # GetTree allows for missing subtrees, but knowing some digests
                # are missing without scanning the result on the caller side
                # makes certain usages more efficient
                if raise_on_missing_subdir and len(blobs) < len(queue):
                    raise NotFoundError(
                        f"Missing entries under root directory: {root_digest.hash}/{root_digest.size_bytes}"
                    )

                directories = [Directory.FromString(b) for b in blobs.values()]
                queue = [subdir.digest for d in directories for subdir in d.directories]

                if len(directories) > 0:
                    yield from directories



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Iterable, Iterator, Union, cast

import grpc

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import DESCRIPTOR as RE_DESCRIPTOR
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ExecuteRequest, WaitExecutionRequest
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2_grpc import (
    ExecutionServicer,
    add_ExecutionServicer_to_server,
)
from buildgrid._protos.google.longrunning import operations_pb2
from buildgrid.server.context import current_instance
from buildgrid.server.decorators import rpc
from buildgrid.server.exceptions import CancelledError
from buildgrid.server.execution.instance import ExecutionInstance
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metadata import extract_client_identity, extract_request_metadata
from buildgrid.server.servicer import InstancedServicer
from buildgrid.server.utils.cancellation import CancellationContext

LOGGER = buildgrid_logger(__name__)


def _parse_instance_name(operation_name: str) -> str:
    names = operation_name.split("/")
    return "/".join(names[:-1])


class ExecutionService(ExecutionServicer, InstancedServicer[ExecutionInstance]):
    SERVICE_NAME = "Execution"
    REGISTER_METHOD = add_ExecutionServicer_to_server
    FULL_NAME = RE_DESCRIPTOR.services_by_name[SERVICE_NAME].full_name

    @rpc(instance_getter=lambda r: cast(str, r.instance_name))
    def Execute(self, request: ExecuteRequest, context: grpc.ServicerContext) -> Iterator[operations_pb2.Operation]:
        yield from self._handle_request(request, context)

    @rpc(instance_getter=lambda r: _parse_instance_name(r.name))
    def WaitExecution(
        self, request: WaitExecutionRequest, context: grpc.ServicerContext
    ) -> Iterator[operations_pb2.Operation]:
        yield from self._handle_request(request, context)

    def query_connected_clients_for_instance(self, instance_name: str) -> int:
        if instance := self.instances.get(instance_name):
            return instance.scheduler.ops_notifier.listener_count()
        return 0

    def _handle_request(
        self,
        request: Union[ExecuteRequest, WaitExecutionRequest],
        context: grpc.ServicerContext,
    ) -> Iterable[operations_pb2.Operation]:
        peer_uid = context.peer()

        try:
            if isinstance(request, ExecuteRequest):
                request_metadata = extract_request_metadata(context.invocation_metadata())
                client_identity = extract_client_identity(current_instance(), context.invocation_metadata())
                operation_name = self.current_instance.execute(
                    action_digest=request.action_digest,
                    skip_cache_lookup=request.skip_cache_lookup,
                    priority=request.execution_policy.priority,
                    request_metadata=request_metadata,
                    client_identity=client_identity,
                )
            else:  # isinstance(request, WaitExecutionRequest)"
                names = request.name.split("/")
                operation_name = names[-1]

            operation_full_name = f"{current_instance()}/{operation_name}"

            for operation in self.current_instance.stream_operation_updates(
                operation_name, context=CancellationContext(context)
            ):
                operation.name = operation_full_name
                yield operation

            if not context.is_active():
                LOGGER.info(
                    "Peer was holding thread for operation updates but the rpc context is inactive; releasing thread.",
                    tags=dict(peer_uid=peer_uid, instance_name=current_instance(), operation_name=operation_name),
                )

        except CancelledError as e:
            LOGGER.info(e)
            context.set_details(str(e))
            context.set_code(grpc.StatusCode.CANCELLED)
            yield e.last_response  # type: ignore[misc]  # need a better signature



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
from typing import Any

import click
from google.protobuf import json_format

from buildgrid.server.client.capabilities import CapabilitiesInterface
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.exceptions import InvalidArgumentError

from ..cli import pass_context


@click.command(name="capabilities", short_help="Capabilities service.")
@click.option(
    "--remote",
    type=click.STRING,
    default="http://localhost:50051",
    show_default=True,
    help="Remote execution server's URL (port defaults to 50051 if no specified).",
)
@click.option(
    "--auth-token",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Authorization token for the remote.",
)
@click.option(
    "--client-key",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Private client key for TLS (PEM-encoded).",
)
@click.option(
    "--client-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public client certificate for TLS (PEM-encoded).",
)
@click.option(
    "--server-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public server certificate for TLS (PEM-encoded).",
)
@click.option(
    "--instance-name", type=click.STRING, default=None, show_default=True, help="Targeted farm instance name."
)
@click.option("--action-id", type=str, help="Action ID.")
@click.option("--invocation-id", type=str, help="Tool invocation ID.")
@click.option("--correlation-id", type=str, help="Correlated invocation ID.")
@pass_context
def cli(
    context: Any,
    remote: str,
    instance_name: str,
    auth_token: str,
    client_key: str,
    client_cert: str,
    server_cert: str,
    action_id: str,
    invocation_id: str,
    correlation_id: str,
) -> None:
    """Entry point for the bgd-capabilities CLI command group."""
    try:
        context.channel, _ = setup_channel(
            remote,
            auth_token=auth_token,
            client_key=client_key,
            client_cert=client_cert,
            server_cert=server_cert,
            action_id=action_id,
            tool_invocation_id=invocation_id,
            correlated_invocations_id=correlation_id,
        )

    except InvalidArgumentError as e:
        click.echo(f"Error: {e}.", err=True)
        sys.exit(-1)

    context.instance_name = instance_name

    interface = CapabilitiesInterface(context.channel)
    try:
        response = interface.get_capabilities(context.instance_name)
    except Exception as e:
        click.echo(f"Error: Getting capabilities: {e}", err=True)
        sys.exit(-1)

    click.echo(json_format.MessageToJson(response))



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License' is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Dict, Protocol, Tuple, Type, TypeVar, Union

GrpcTrailingMetadata = Tuple[Tuple[str, Union[str, bytes]], ...]
LogStreamConnectionConfig = Dict[str, Union[str, Dict[str, str]]]


class OnServerStartCallback(Protocol):
    def __call__(self) -> Any: ...


class PortAssignedCallback(Protocol):
    def __call__(self, port_map: Dict[Any, int]) -> Any: ...


Cls = TypeVar("Cls")


class MessageType(Protocol):
    # Acts like a protobuf message. Add methods as needed.

    @classmethod
    def FromString(cls: Type[Cls], s: bytes) -> Cls:
        pass

    def SerializeToString(self: Any, *, deterministic: bool = ...) -> bytes:
        pass

    def ParseFromString(self, serialized: bytes) -> int:
        pass

    def Clear(self) -> None:
        pass



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from contextvars import ContextVar
from typing import Iterator, Optional

from buildgrid.server.exceptions import InvalidArgumentError

_instance_name: "ContextVar[Optional[str]]" = ContextVar("_instance_name", default=None)
_service_name: "ContextVar[Optional[str]]" = ContextVar("_service_name", default=None)
_method_name: "ContextVar[Optional[str]]" = ContextVar("_method_name", default=None)


@contextmanager
def instance_context(instance_name: Optional[str]) -> Iterator[None]:
    token = _instance_name.set(instance_name)
    try:
        yield
    finally:
        _instance_name.reset(token)


def try_current_instance() -> Optional[str]:
    return _instance_name.get()


def current_instance() -> str:
    instance_name = try_current_instance()
    if instance_name is None:
        raise InvalidArgumentError("Instance name not set")
    return instance_name


@contextmanager
def service_context(service_name: Optional[str]) -> Iterator[None]:
    token = _service_name.set(service_name)
    try:
        yield
    finally:
        _service_name.reset(token)


def try_current_service() -> Optional[str]:
    return _service_name.get()


def current_service() -> str:
    service_name = try_current_service()
    if service_name is None:
        raise InvalidArgumentError("Service name not set")
    return service_name


@contextmanager
def method_context(method_name: Optional[str]) -> Iterator[None]:
    token = _method_name.set(method_name)
    try:
        yield
    finally:
        _method_name.reset(token)


def try_current_method() -> Optional[str]:
    return _method_name.get()


def current_method() -> str:
    method_name = try_current_method()
    if method_name is None:
        raise InvalidArgumentError("Method name not set")
    return method_name



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from contextlib import ExitStack
from typing import Any, Dict, List, Optional, Tuple

from buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2 import DESCRIPTOR as BOTS_DESCRIPTOR
from buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2 import BotSession, Lease
from buildgrid.server.context import current_instance
from buildgrid.server.enums import BotStatus
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.scheduler import Scheduler
from buildgrid.server.servicer import Instance
from buildgrid.server.settings import MAX_WORKER_TTL, NETWORK_TIMEOUT
from buildgrid.server.utils.cancellation import CancellationContext

LOGGER = buildgrid_logger(__name__)


class BotsInterface(Instance):
    SERVICE_NAME = BOTS_DESCRIPTOR.services_by_name["Bots"].full_name

    def __init__(self, scheduler: Scheduler) -> None:
        self._stack = ExitStack()
        self.scheduler = scheduler

    def start(self) -> None:
        self._stack.enter_context(self.scheduler)
        self._stack.enter_context(self.scheduler.job_assigner)
        self._stack.enter_context(self.scheduler.session_expiry_timer)

    def stop(self) -> None:
        self._stack.close()
        LOGGER.info("Stopped Bots.")

    def create_bot_session(
        self, bot_session: BotSession, context: CancellationContext, deadline: Optional[float] = None
    ) -> BotSession:
        """Creates a new bot session. Server should assign a unique
        name to the session. If the bot_id already exists in the database
        then any leases already assigned to that id are requeued
        (via close_bot_session) and then the name previously associated with
        the bot_id is replaced with the new name in the database. If the bot_id
        is not in the database, a new record is created.
        """
        if not bot_session.bot_id:
            raise InvalidArgumentError("Bot's id must be set by client.")

        # Create new record
        bot_session.name = self.scheduler.add_bot_entry(
            bot_session_id=bot_session.bot_id, bot_session_status=bot_session.status
        )

        LOGGER.info("Created new BotSession. Requesting leases.", tags=self._bot_log_tags(bot_session))
        self._request_leases(bot_session, context, deadline=deadline)
        self._assign_deadline_for_botsession(bot_session)

        LOGGER.debug("Completed CreateBotSession.", tags=self._bot_log_tags(bot_session))
        return bot_session

    def update_bot_session(
        self, bot_session: BotSession, context: CancellationContext, deadline: Optional[float] = None
    ) -> Tuple[BotSession, List[Tuple[str, bytes]]]:
        """Client updates the server. Any changes in state to the Lease should be
        registered server side. Assigns available leases with work.
        """
        LOGGER.debug("Beginning initial lease synchronization.", tags=self._bot_log_tags(bot_session))

        orig_lease: Optional[Lease] = None
        if bot_session.leases:
            orig_lease = bot_session.leases.pop()

        if updated_lease := self.scheduler.synchronize_bot_lease(
            bot_session.name, bot_session.bot_id, bot_session.status, orig_lease
        ):
            bot_session.leases.append(updated_lease)

        LOGGER.debug("Completed initial lease synchronization.", tags=self._bot_log_tags(bot_session))

        # Don't request new leases if a lease was removed. This mitigates situations where the scheduler
        # is updated with the new state of the lease, but a fault thereafter causes the worker to retry
        # the old UpdateBotSession call
        if not orig_lease and not updated_lease:
            self._request_leases(bot_session, context, deadline=deadline)

        metadata = self.scheduler.get_metadata_for_leases(bot_session.leases)
        self._assign_deadline_for_botsession(bot_session)

        LOGGER.debug("Completed UpdateBotSession.", tags=self._bot_log_tags(bot_session))
        return bot_session, metadata

    def count_bots_by_status(self) -> Dict[BotStatus, int]:
        return self.scheduler.count_bots_by_status()

    def _assign_deadline_for_botsession(self, bot_session: BotSession) -> None:
        bot_session.expire_time.FromDatetime(
            self.scheduler.refresh_bot_expiry_time(bot_session.name, bot_session.bot_id)
        )

    def _request_leases(
        self,
        bot_session: BotSession,
        context: CancellationContext,
        deadline: Optional[float] = None,
    ) -> None:
        # We do not assign new leases if we are not in the OK state.
        if bot_session.status != BotStatus.OK.value:
            LOGGER.debug("BotSession not healthy. Skipping lease request.", tags=self._bot_log_tags(bot_session))
            return

        # Only send one lease at a time currently. If any leases are set we can abort the request.
        if bot_session.leases:
            LOGGER.debug("BotSession already assigned. Skipping lease request.", tags=self._bot_log_tags(bot_session))
            return

        # If no deadline is set default to the max we allow workers to long-poll for work
        if deadline is None:
            deadline = MAX_WORKER_TTL

        # If the specified bot session keepalive timeout is greater than the
        # deadline it can result in active bot sessions being reaped
        deadline = min(deadline, self.scheduler.bot_session_keepalive_timeout)

        # Use 80% of the given deadline to give time to respond but no less than NETWORK_TIMEOUT
        ttl = deadline * 0.8
        if ttl < NETWORK_TIMEOUT:
            LOGGER.info(
                "BotSession expires in less time than timeout. No leases will be assigned.",
                tags={**self._bot_log_tags(bot_session), "network_timeout": NETWORK_TIMEOUT},
            )
            return

        # Wait for an update to the bot session and then resynchronize the lease.
        LOGGER.debug("Waiting for job assignment.", tags={**self._bot_log_tags(bot_session), "deadline": deadline})
        with self.scheduler.job_assigner.assignment_context(bot_session) as event:
            context.on_cancel(event.set)
            event.wait(ttl)

        # This is a best-effort check the see if the original request is still alive. Depending on
        # network and proxy configurations, this status may not accurately reflect the state of the
        # client connection. If we know for certain that the request is no longer being monitored,
        # we can exit now to avoid state changes not being acked by the bot.
        if context.is_cancelled():
            LOGGER.debug(
                "Bot request cancelled. Skipping lease synchronization.", tags=self._bot_log_tags(bot_session)
            )
            return

        # In the case that we had a timeout, we can return without post lease synchronization. This
        # helps deal with the case of uncommunicated cancellations from the bot request. If the bot
        # is actually still waiting on work, this will be immediately followed up by a new request
        # from the worker, where the initial synchronization will begin a bot ack for the pending
        # job. In the case that the request has been abandoned, it avoids competing updates to the
        # database records in the corresponding bots session.
        if not event.is_set():
            LOGGER.debug(
                "Bot assignment timeout. Skipping lease synchronization.", tags=self._bot_log_tags(bot_session)
            )
            return

        # Synchronize the lease again to pick up db changes.
        LOGGER.debug("Synchronizing leases after job assignment wait.", tags=self._bot_log_tags(bot_session))
        if lease := self.scheduler.synchronize_bot_lease(
            bot_session.name, bot_session.bot_id, bot_session.status, None
        ):
            bot_session.leases.append(lease)

    def _bot_log_tags(self, bot_session: BotSession) -> Dict[str, Any]:
        lease_id, lease_state = None, None
        if bot_session.leases:
            lease_id, lease_state = bot_session.leases[0].id, bot_session.leases[0].state
        return {
            "instance_name": current_instance(),
            "request.bot_name": bot_session.name,
            "request.bot_id": bot_session.bot_id,
            "request.bot_status": bot_session.status,
            "request.lease_id": lease_id,
            "request.lease_state": lease_state,
        }



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
OperationsService
=================

"""

from typing import cast

import grpc
from google.protobuf.empty_pb2 import Empty

from buildgrid._protos.google.longrunning.operations_pb2 import DESCRIPTOR as OPS_DESCRIPTOR
from buildgrid._protos.google.longrunning.operations_pb2 import (
    CancelOperationRequest,
    DeleteOperationRequest,
    GetOperationRequest,
    ListOperationsRequest,
    ListOperationsResponse,
    Operation,
)
from buildgrid._protos.google.longrunning.operations_pb2_grpc import (
    OperationsServicer,
    add_OperationsServicer_to_server,
)
from buildgrid.server.decorators import rpc
from buildgrid.server.operations.instance import OperationsInstance
from buildgrid.server.servicer import InstancedServicer
from buildgrid.server.settings import CLIENT_IDENTITY_HEADER_NAME, REQUEST_METADATA_HEADER_NAME


def _parse_instance_name(operation_name: str) -> str:
    names = operation_name.split("/")
    return "/".join(names[:-1]) if len(names) > 1 else ""


def _parse_operation_name(name: str) -> str:
    names = name.split("/")
    return names[-1] if len(names) > 1 else name


class OperationsService(OperationsServicer, InstancedServicer[OperationsInstance]):
    SERVICE_NAME = "Operations"
    REGISTER_METHOD = add_OperationsServicer_to_server
    FULL_NAME = OPS_DESCRIPTOR.services_by_name[SERVICE_NAME].full_name

    @rpc(instance_getter=lambda r: _parse_instance_name(r.name))
    def GetOperation(self, request: GetOperationRequest, context: grpc.ServicerContext) -> Operation:
        operation_name = _parse_operation_name(request.name)
        operation, metadata, client_identity = self.current_instance.get_operation(operation_name)
        operation.name = request.name

        trailing_metadata = []
        if metadata is not None:
            trailing_metadata.append((REQUEST_METADATA_HEADER_NAME, metadata.SerializeToString()))
        if client_identity is not None:
            trailing_metadata.append((CLIENT_IDENTITY_HEADER_NAME, client_identity.SerializeToString()))

        if trailing_metadata:
            context.set_trailing_metadata(trailing_metadata)  # type: ignore[arg-type]  # tricky covariance issue

        return operation

    @rpc(instance_getter=lambda r: cast(str, r.name))
    def ListOperations(self, request: ListOperationsRequest, context: grpc.ServicerContext) -> ListOperationsResponse:
        # The request name should be the collection name. In our case, this is just the instance name
        result = self.current_instance.list_operations(request.filter, request.page_size, request.page_token)

        for operation in result.operations:
            operation.name = f"{request.name}/{operation.name}"

        return result

    @rpc(instance_getter=lambda r: _parse_instance_name(r.name))
    def DeleteOperation(self, request: DeleteOperationRequest, context: grpc.ServicerContext) -> Empty:
        context.set_details("BuildGrid does not support DeleteOperation.")
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        return Empty()

    @rpc(instance_getter=lambda r: _parse_instance_name(r.name))
    def CancelOperation(self, request: CancelOperationRequest, context: grpc.ServicerContext) -> Empty:
        self.current_instance.cancel_operation(_parse_operation_name(request.name))
        return Empty()



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Server command
=================

Create a BuildGrid server.
"""

import functools
import os
import signal
import sys
from typing import Any, Dict, Optional

import click
from jsonschema.exceptions import ValidationError

from buildgrid.server.app.cli import Context, pass_context, setup_logging
from buildgrid.server.app.settings.config import populate_buildgrid_config
from buildgrid.server.app.settings.parser import load_config, validate_config
from buildgrid.server.auth.manager import set_auth_manager
from buildgrid.server.exceptions import PermissionDeniedError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.monitoring import set_monitoring_bus
from buildgrid.server.server import Server

LOGGER = buildgrid_logger(__name__)


@click.group(name="server", short_help="Start a local server instance.")
@pass_context
def cli(context: Context) -> None:
    pass


@cli.command(name="lint", short_help="Check the format of a buildgrid server configuration.")
@click.option("-v", "--verbose", count=True, help="Increase log verbosity level.")
@click.argument("CONFIG", type=click.Path(file_okay=True, dir_okay=False, exists=True, writable=False))
@pass_context
def lint(context: Context, config: "os.PathLike[str]", verbose: int) -> None:
    setup_logging(verbosity=verbose)

    try:
        validate_config(config, strict=True)
    except ValidationError as e:
        click.echo(click.style(f"ERROR: Config ({config}) failed validation: {e}", fg="red", bold=True), err=True)
        sys.exit(-1)

    click.echo(click.style(f"Success: Config ({config}) passed validation", fg="green", bold=True))


@cli.command("start", short_help="Setup a new server instance.")
@click.argument("CONFIG", type=click.Path(file_okay=True, dir_okay=False, exists=True, writable=False))
@click.option("-v", "--verbose", count=True, help="Increase log verbosity level.")
@click.option("--pid-file", type=click.Path(dir_okay=False), help="Path to PID file")
@pass_context
def start(context: Context, config: "os.PathLike[str]", verbose: int, pid_file: "os.PathLike[str]") -> None:
    """Entry point for the bgd-server CLI command group."""
    setup_logging(verbosity=verbose)

    click.echo(f"\nLoading config from {config}")

    try:
        settings = load_config(config)
    except ValidationError as e:
        click.echo(click.style(f"ERROR: Config ({config}) failed validation: {e}", fg="red", bold=True), err=True)
        sys.exit(-1)

    try:
        server = _create_server_from_config(settings)

    except KeyError as e:
        click.echo(f"ERROR: Could not parse config: {e}.\n", err=True)
        sys.exit(-1)

    signal.signal(signal.SIGTERM, server.stop)
    signal.signal(signal.SIGINT, server.stop)
    on_server_start_cb = functools.partial(_create_new_pid_file, pid_file)
    click.echo(click.style("Starting BuildGrid server...", fg="green", bold=True))
    try:
        server.start(on_server_start_cb=on_server_start_cb)
    except KeyboardInterrupt:
        pass
    finally:
        server.stop()
        _remove_old_pid_file(pid_file)


def _remove_old_pid_file(pid_file: Optional["os.PathLike[str]"]) -> None:
    """Remove pid_file if it's set"""
    if not pid_file:
        return

    try:
        os.remove(pid_file)
    except os.error:
        LOGGER.error("Error deleting pid-file.", tags=dict(pid_file=str(pid_file)), exc_info=True)


def _create_new_pid_file(pid_file: Optional["os.PathLike[str]"]) -> None:
    if pid_file:
        with open(pid_file, "w", encoding="utf-8") as f:
            f.write(str(os.getpid()))

        LOGGER.info("Created pid-file.", tags=dict(pid_file=str(pid_file)))


def _create_server_from_config(raw_configuration: Dict[str, Any]) -> Server:
    """Parses configuration and setup a fresh server instance."""

    try:
        config = populate_buildgrid_config(raw_configuration)
    except ValueError as e:
        click.echo(f"Error: populating server config: {e}.", err=True)
        sys.exit(-1)

    if any(instance.name == "unnamed" for instance in config.instances):
        click.echo("Error: The instance name 'unnamed' is reserved to avoid metrics collisions", err=True)
        sys.exit(-1)

    server = Server(
        server_reflection=config.server_reflection,
        grpc_compression=config.grpc_compression,
        is_instrumented=config.monitoring is not None,
        grpc_server_options=config.grpc_server_options,
        max_workers=config.thread_pool_size,
    )

    if monitoring_bus := config.monitoring:
        set_monitoring_bus(monitoring_bus)

    if auth := config.authorization:
        set_auth_manager(auth)

    try:
        for channel in config.server:
            server.add_port(channel.address, channel.credentials)

    except PermissionDeniedError as e:
        click.echo(f"Error: {e}.", err=True)
        sys.exit(-1)

    for instance in config.instances:
        for service in instance.services:
            server.register_instance(instance.name, service)

    return server



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from cachetools import TTLCache

from buildgrid.server.exceptions import InvalidArgumentError

DEFAULT_TOKEN_REFRESH_SECONDS = 60 * 60
_CACHE_KEY = "token"


class AuthTokenLoader:

    def __init__(self, token_path: str, refresh_in_seconds: Optional[int] = None) -> None:
        self._token_path = token_path
        self._cache: TTLCache[str, str] = TTLCache(
            maxsize=1, ttl=refresh_in_seconds if refresh_in_seconds else DEFAULT_TOKEN_REFRESH_SECONDS
        )

    def _load_token_from_file(self) -> None:
        try:
            with open(self._token_path, mode="r") as token_file:
                token = token_file.read().strip()
                self._cache[_CACHE_KEY] = token
        except Exception as e:
            raise InvalidArgumentError(f"Cannot read token from filepath: {self._token_path}") from e

    def get_token(self) -> str:
        if _CACHE_KEY not in self._cache:
            self._load_token_from_file()

        assert _CACHE_KEY in self._cache
        return self._cache[_CACHE_KEY]



import random
import threading
import uuid
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional

from buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2 import BotSession
from buildgrid.server.context import current_instance, instance_context
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.threading import ContextWorker

from .properties import PropertySet, hash_from_dict

if TYPE_CHECKING:
    # Avoid circular import
    from .impl import Scheduler


LOGGER = buildgrid_logger(__name__)


class JobAssigner:
    def __init__(
        self,
        scheduler: "Scheduler",
        property_set: PropertySet,
        job_assignment_interval: float = 1.0,
        priority_percentage: int = 100,
    ):
        self._lock = threading.Lock()
        # Dict[Instance, Dict[Hash, Dict[BotName, Dict[Key, Event]]]]
        self._events: Dict[str, Dict[str, Dict[str, Dict[str, threading.Event]]]] = {}
        self._scheduler = scheduler
        self._property_set = property_set
        # Here we allow immediately starting a new assignment if a bot is added to the lookup.
        self._new_bots_added = threading.Event()
        self.assigner = ContextWorker(
            target=self.begin, name="JobAssignment", on_shutdown_requested=self._new_bots_added.set
        )
        self.job_assignment_interval = job_assignment_interval
        self.priority_percentage = priority_percentage

    def __enter__(self) -> "JobAssigner":
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()

    def start(self) -> None:
        self.assigner.start()

    def stop(self) -> None:
        self.assigner.stop()

    def listener_count(self, instance_name: Optional[str] = None) -> int:
        with self._lock:
            return len(
                {
                    bot_name
                    for event_instance_name, instance_events in self._events.items()
                    if instance_name is None or instance_name == event_instance_name
                    for bot_events in instance_events.values()
                    for bot_name in bot_events
                }
            )

    @contextmanager
    def assignment_context(self, bot_session: BotSession) -> Iterator[threading.Event]:
        key = str(uuid.uuid4())
        event = threading.Event()
        worker_hashes = set(map(hash_from_dict, self._property_set.worker_properties(bot_session)))
        instance_name = current_instance()
        try:
            with self._lock:
                self._events.setdefault(instance_name, {})
                for worker_hash in worker_hashes:
                    self._events[instance_name].setdefault(worker_hash, {})
                    self._events[instance_name][worker_hash].setdefault(bot_session.name, {})
                    self._events[instance_name][worker_hash][bot_session.name][key] = event
            self._new_bots_added.set()
            yield event
        finally:
            with self._lock:
                for worker_hash in worker_hashes:
                    del self._events[instance_name][worker_hash][bot_session.name][key]
                    if len(self._events[instance_name][worker_hash][bot_session.name]) == 0:
                        del self._events[instance_name][worker_hash][bot_session.name]
                    if len(self._events[instance_name][worker_hash]) == 0:
                        del self._events[instance_name][worker_hash]
                if len(self._events[instance_name]) == 0:
                    del self._events[instance_name]

    def assign_jobs(self, shutdown_requested: threading.Event, instance_name: str, oldest_first: bool = False) -> None:
        """Assign jobs to the currently connected workers

        This method iterates over the buckets of currently connected workers,
        and requests a number of job assignments from the scheduler to cover
        the number of workers in each bucket. Empty buckets are skipped.
        """

        with self._lock:
            worker_hashes = list(self._events.get(instance_name, {}).keys())

        random.shuffle(worker_hashes)
        for worker_hash in worker_hashes:
            if shutdown_requested.is_set():
                return

            with self._lock:
                bot_names = list(self._events.get(instance_name, {}).get(worker_hash, {}))

            if bot_names:
                if oldest_first:
                    assigned_bot_names = self._scheduler.assign_n_leases_by_age(
                        capability_hash=worker_hash, bot_names=bot_names
                    )
                else:
                    assigned_bot_names = self._scheduler.assign_n_leases_by_priority(
                        capability_hash=worker_hash, bot_names=bot_names
                    )
                with self._lock:
                    for name in assigned_bot_names:
                        for event in self._events.get(instance_name, {}).get(worker_hash, {}).get(name, {}).values():
                            event.set()

    def begin(self, shutdown_requested: threading.Event) -> None:
        while not shutdown_requested.is_set():
            oldest_first = random.randint(1, 100) > self.priority_percentage

            with self._lock:
                instance_names = list(self._events)

            for instance_name in instance_names:
                try:
                    with instance_context(instance_name):
                        self.assign_jobs(shutdown_requested, instance_name, oldest_first=oldest_first)
                except Exception:
                    LOGGER.exception(
                        "Error in job assignment thread.", tags=dict(instance_name=instance_name), exc_info=True
                    )

            self._new_bots_added.wait(timeout=self.job_assignment_interval)
            self._new_bots_added.clear()



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Store tool information for a job in the database


Revision ID: 29e88b3e0d0a
Revises: 8d910c8de8b6
Create Date: 2020-10-16 14:35:49.304924

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "29e88b3e0d0a"
down_revision = "8d910c8de8b6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("operations", sa.Column("correlated_invocations_id", sa.String(), nullable=True))
    op.add_column("operations", sa.Column("invocation_id", sa.String(), nullable=True))
    op.add_column("operations", sa.Column("tool_name", sa.String(), nullable=True))
    op.add_column("operations", sa.Column("tool_version", sa.String(), nullable=True))


def downgrade() -> None:
    op.drop_column("operations", "tool_version")
    op.drop_column("operations", "tool_name")
    op.drop_column("operations", "invocation_id")
    op.drop_column("operations", "correlated_invocations_id")



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add blobs table for SQLStorage

Revision ID: 316ad77858af
Revises: d4efbcc698ca
Create Date: 2024-02-20 12:39:33.255751

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "316ad77858af"
down_revision = "d4efbcc698ca"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "blobs",
        sa.Column("digest_hash", sa.String(), nullable=False),
        sa.Column("digest_size_bytes", sa.BigInteger(), nullable=False),
        sa.Column("data", sa.LargeBinary(), nullable=False),
        sa.PrimaryKeyConstraint("digest_hash"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("blobs")
    # ### end Alembic commands ###



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import DESCRIPTOR as RE_DESCRIPTOR
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_distribution_metric
from buildgrid.server.servicer import Instance

LOGGER = buildgrid_logger(__name__)
EPOCH = datetime(1970, 1, 1)


class ActionCache(Instance):
    SERVICE_NAME = RE_DESCRIPTOR.services_by_name["ActionCache"].full_name

    def __init__(self, cache: ActionCacheABC) -> None:
        """Initialises a new ActionCache instance.

        Args:
            cache (ActionCacheABC): The cache to use to store results.

        """
        self._cache = cache

    # --- Public API ---

    def start(self) -> None:
        self._cache.start()

    def stop(self) -> None:
        self._cache.stop()
        LOGGER.info("Stopped ActionCache.")

    @property
    def allow_updates(self) -> bool:
        return self._cache.allow_updates

    def get_action_result(self, action_digest: Digest) -> ActionResult:
        """Retrieves the cached result for an Action.

        If there is no cached result found, returns None.

        Args:
            action_digest (Digest): The digest of the Action to retrieve the
                cached result of.

        """
        res = self._cache.get_action_result(action_digest)

        time_completed = res.execution_metadata.worker_completed_timestamp.ToDatetime()
        if time_completed != EPOCH:
            age = datetime.now() - time_completed
            age_in_ms = age / timedelta(milliseconds=1)
            publish_distribution_metric(METRIC.ACTION_CACHE.RESULT_AGE, age_in_ms)

        return res

    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        """Stores a result for an Action in the cache.

        If the result has a non-zero exit code and `cache_failed_actions` is False
        for this cache, the result is not cached.

        Args:
            action_digest (Digest): The digest of the Action whose result is
                being cached.
            action_result (ActionResult): The result to cache for the given
                Action digest.

        """
        self._cache.update_action_result(action_digest, action_result)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from buildgrid._protos.google.devtools.remoteworkers.v1test2 import tasks_pb2 as google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2


class TasksStub(object):
    """DEPRECATED. GetTask should be replaced by Lease.payload, UpdateTaskResult by
    Lease.result and logs should be precreated prior to sending to the bot (eg,
    via CommandTask.expected_outputs.stdout_destination).
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetTask = channel.unary_unary(
                '/google.devtools.remoteworkers.v1test2.Tasks/GetTask',
                request_serializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.GetTaskRequest.SerializeToString,
                response_deserializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.Task.FromString,
                )
        self.UpdateTaskResult = channel.unary_unary(
                '/google.devtools.remoteworkers.v1test2.Tasks/UpdateTaskResult',
                request_serializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.UpdateTaskResultRequest.SerializeToString,
                response_deserializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.TaskResult.FromString,
                )
        self.AddTaskLog = channel.unary_unary(
                '/google.devtools.remoteworkers.v1test2.Tasks/AddTaskLog',
                request_serializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.AddTaskLogRequest.SerializeToString,
                response_deserializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.AddTaskLogResponse.FromString,
                )


class TasksServicer(object):
    """DEPRECATED. GetTask should be replaced by Lease.payload, UpdateTaskResult by
    Lease.result and logs should be precreated prior to sending to the bot (eg,
    via CommandTask.expected_outputs.stdout_destination).
    """

    def GetTask(self, request, context):
        """DEPRECATED - use Lease.payload instead.
        GetTask reads the current state of the task. Tasks must be created through
        some other interface, and should be immutable once created and exposed to
        the bots.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateTaskResult(self, request, context):
        """DEPRECATED - use Lease.result instead.
        UpdateTaskResult updates the result.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddTaskLog(self, request, context):
        """DEPRECATED - precreate logs prior to sending to bot.
        AddTaskLog creates a new streaming log. The log is streamed and marked as
        completed through other interfaces (i.e., ByteStream). This can be called
        by the bot if it wants to create a new log; the server can also predefine
        logs that do not need to be created (e.g. `stdout`).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TasksServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetTask': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTask,
                    request_deserializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.GetTaskRequest.FromString,
                    response_serializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.Task.SerializeToString,
            ),
            'UpdateTaskResult': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateTaskResult,
                    request_deserializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.UpdateTaskResultRequest.FromString,
                    response_serializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.TaskResult.SerializeToString,
            ),
            'AddTaskLog': grpc.unary_unary_rpc_method_handler(
                    servicer.AddTaskLog,
                    request_deserializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.AddTaskLogRequest.FromString,
                    response_serializer=google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.AddTaskLogResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'google.devtools.remoteworkers.v1test2.Tasks', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Tasks(object):
    """DEPRECATED. GetTask should be replaced by Lease.payload, UpdateTaskResult by
    Lease.result and logs should be precreated prior to sending to the bot (eg,
    via CommandTask.expected_outputs.stdout_destination).
    """

    @staticmethod
    def GetTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/google.devtools.remoteworkers.v1test2.Tasks/GetTask',
            google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.GetTaskRequest.SerializeToString,
            google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.Task.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateTaskResult(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/google.devtools.remoteworkers.v1test2.Tasks/UpdateTaskResult',
            google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.UpdateTaskResultRequest.SerializeToString,
            google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.TaskResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddTaskLog(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/google.devtools.remoteworkers.v1test2.Tasks/AddTaskLog',
            google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.AddTaskLogRequest.SerializeToString,
            google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_tasks__pb2.AddTaskLogResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import cast

import grpc

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import DESCRIPTOR as RE_DESCRIPTOR
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    GetCapabilitiesRequest,
    ServerCapabilities,
)
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2_grpc import (
    CapabilitiesServicer,
    add_CapabilitiesServicer_to_server,
)
from buildgrid.server.capabilities.instance import CapabilitiesInstance
from buildgrid.server.decorators import rpc
from buildgrid.server.servicer import InstancedServicer


class CapabilitiesService(CapabilitiesServicer, InstancedServicer[CapabilitiesInstance]):
    SERVICE_NAME = "Capabilities"
    REGISTER_METHOD = add_CapabilitiesServicer_to_server
    FULL_NAME = RE_DESCRIPTOR.services_by_name[SERVICE_NAME].full_name

    @property
    def enabled(self) -> bool:
        # We always want a capabilities service
        return True

    @rpc(instance_getter=lambda r: cast(str, r.instance_name))
    def GetCapabilities(self, request: GetCapabilitiesRequest, context: grpc.ServicerContext) -> ServerCapabilities:
        return self.current_instance.get_capabilities()
